from decimal import Decimal
from django import forms
from django.conf import settings
from django.template import loader
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _
from l10n.utils import moneyfmt
from livesettings import config_value, config_value_safe, config_choice_values, config_get_group
from payment import signals
from payment.config import labelled_gateway_choices
from payment.models import CreditCardDetail
from payment.utils import get_or_create_order, pay_ship_save
from product.models import Discount, TaxClass, Price, PriceAdjustmentCalc, PriceAdjustment
from product.utils import find_best_auto_discount
from satchmo_store.contact.forms import ProxyContactForm, ContactInfoForm
from satchmo_store.contact.models import Contact
from satchmo_store.shop.models import Cart, Order
from satchmo_store.shop.signals import satchmo_shipping_price_query
from satchmo_utils.dynamic import lookup_template
from satchmo_utils.views import CreditCard
from shipping.config import shipping_methods, shipping_method_by_key
from shipping.signals import shipping_choices_query
from shipping.utils import update_shipping
from signals_ahoy.signals import form_init, form_initialdata, form_presave, form_postsave, form_validate
from tax.templatetags.satchmo_tax import _get_taxprocessor
from threaded_multihost import threadlocals
import calendar
import datetime
import logging
import sys

log = logging.getLogger('payment.forms')

MONTHS = [(month,'%02d'%month) for month in range(1,13)]

def _get_cheapest_shipping(shipping_dict):
    """Use the shipping_dict as returned by _get_shipping_choices
    to figure the cheapest shipping option."""
    
    least = None
    leastcost = None
    for key, value in shipping_dict.items():
        current = value['cost']
        if leastcost is None or current < leastcost:
            least = key
            leastcost = current

    return least

def _get_shipping_choices(request, paymentmodule, cart, contact, default_view_tax=False, order=None):
    """Iterate through legal shipping modules, building the list for display to the user.
    
    Returns the shipping choices list, along with a dictionary of shipping choices, useful
    for building javascript that operates on shipping choices.
    """
    shipping_options = []
    shipping_dict = {}
    rendered = {}
    if not order:
        try:
            order = Order.objects.from_request(request)
        except Order.DoesNotExist:
            pass
    
    discount = None
    if order:
        try:
            discount = Discount.objects.by_code(order.discount_code)
        except Discount.DoesNotExist:
            pass

    if not cart.is_shippable:
        methods = [shipping_method_by_key('NoShipping'),]
    else:
        methods = shipping_methods()
        
    tax_shipping = config_value_safe('TAX','TAX_SHIPPING', False)
    shipping_tax = None
    
    if tax_shipping:
        taxer = _get_taxprocessor(request)
        shipping_tax = TaxClass.objects.get(title=config_value('TAX', 'TAX_CLASS'))
    
    for method in methods:
        method.calculate(cart, contact)
        if method.valid():
            template = lookup_template(paymentmodule, 'shipping/options.html')
            t = loader.get_template(template)
            shipcost = finalcost = method.cost()

            if discount and order:
                order.shipping_cost = shipcost
                discount.calc(order)
                shipdiscount = discount.item_discounts.get('Shipping', 0)
            else:
                shipdiscount = 0
            
            # set up query to determine shipping price to show
            shipprice = Price()
            shipprice.price = shipcost
            shipadjust = PriceAdjustmentCalc(shipprice)
            if shipdiscount:
                shipadjust += PriceAdjustment('discount', _('Discount'), shipdiscount)

            satchmo_shipping_price_query.send(cart, adjustment=shipadjust)
            shipdiscount = shipadjust.total_adjustment()

            if shipdiscount:
                finalcost -= shipdiscount

            shipping_dict[method.id] = {'cost' : shipcost, 'discount' : shipdiscount, 'final' : finalcost}

            taxed_shipping_price = None
            if tax_shipping:
                taxcost = taxer.by_price(shipping_tax, finalcost)
                total = finalcost + taxcost
                taxed_shipping_price = moneyfmt(total)
                shipping_dict[method.id]['taxedcost'] = total
                shipping_dict[method.id]['tax'] = taxcost
                
            c = RequestContext(request, {
                'amount': finalcost,
                'description' : method.description(),
                'method' : method.method(),
                'expected_delivery' : method.expectedDelivery(),
                'default_view_tax' : default_view_tax,
                'shipping_tax': shipping_tax,
                'taxed_shipping_price': taxed_shipping_price})
            rendered[method.id] = t.render(c)
            
    #now sort by price, low to high
    sortme = [(value['cost'], key) for key, value in shipping_dict.items()]
    sortme.sort()
    
    shipping_options = [(key, rendered[key]) for cost, key in sortme]

    shipping_choices_query.send(sender=cart, cart=cart, 
        paymentmodule=paymentmodule, contact=contact, 
        default_view_tax=default_view_tax, order=order,
        shipping_options = shipping_options,
        shipping_dict = shipping_dict)
    return shipping_options, shipping_dict


def _find_sale(cart):
    if cart.numItems > 0:
        products = [item.product for item in cart.cartitem_set.all()]
        sale = find_best_auto_discount(products)
    else:
        sale = None
        
    return sale

class CustomChargeForm(forms.Form):
    orderitem = forms.IntegerField(required=True, widget=forms.HiddenInput())
    amount = forms.DecimalField(label=_('New price'), required=False)
    shipping = forms.DecimalField(label=_('Shipping adjustment'), required=False)
    notes = forms.CharField(_("Notes"), required=False, initial="Your custom item is ready.")
        
    def __init__(self, *args, **kwargs):
        initial = kwargs.get('initial', {})
        form_initialdata.send('CustomChargeForm', form=self, initial=initial)
        kwargs['initial'] = initial
        super(CustomChargeForm, self).__init__(*args, **kwargs)
        form_init.send(self.__class__, form=self)

    def clean(self, *args, **kwargs):
        super(CustomChargeForm, self).clean(*args, **kwargs)
        form_validate.send(self.__class__, form=self)
        return self.cleaned_data

class PaymentMethodForm(ProxyContactForm):
    paymentmethod = forms.ChoiceField(
            label=_('Payment method'),
            choices=labelled_gateway_choices(),
            widget=forms.RadioSelect,
            required=True
            )

    def __init__(self, cart=None, order=None, *args, **kwargs):
        super(PaymentMethodForm, self).__init__(*args, **kwargs)
        self.cart = cart
        # Send a signal to perform additional filtering of available payment methods.
        # Receivers have cart/order passed in variables to check the contents and modify methods
        # list if neccessary.
        payment_choices = labelled_gateway_choices()
        signals.payment_methods_query.send(
                PaymentMethodForm,
                methods=payment_choices,
                cart=cart,
                order=order,
                contact=self._contact
                )
        if len(payment_choices) == 1:
            self.fields['paymentmethod'].widget = forms.HiddenInput(attrs={'value' : payment_choices[0][0]})
        else:
            self.fields['paymentmethod'].widget = forms.RadioSelect(attrs={'value' : payment_choices[0][0]})
        self.fields['paymentmethod'].choices = payment_choices

    def clean(self):
        # allow additional validation
        form_validate.send(PaymentMethodForm, form=self)
        return self.cleaned_data

class PaymentContactInfoForm(PaymentMethodForm, ContactInfoForm):
        payment_required_fields = None

        def __init__(self, *args, **kwargs):
            super(PaymentContactInfoForm, self).__init__(*args, **kwargs)
            if not self.cart:
                request = threadlocals.get_current_request()
                self.cart = Cart.objects.from_request(request)
                
            self.fields['discount'] = forms.CharField(max_length=30, required=False)
            
            self.payment_required_fields = {}

            if config_value('PAYMENT', 'USE_DISCOUNTS'):
                if not self.fields['discount'].initial:
                    sale = _find_sale(self.cart)
                    if sale:
                        self.fields['discount'].initial = sale.code
            else:
                self.fields['discount'].widget = forms.HiddenInput()

            # Listeners of the form_init signal (below) may modify the dict of
            # payment_required_fields. For example, if your CUSTOM_PAYMENT requires
            # customer's city, put the following code in the listener:
            #
            #   form.payment_required_fields['CUSTOM_PAYMENT'] = ['city']
            #
            form_init.send(PaymentContactInfoForm, form=self)

        def save(self, request, *args, **kwargs):
            contactid = super(PaymentContactInfoForm, self).save(*args, **kwargs)
            form_presave.send(PaymentContactInfoForm, form=self)
            contact = Contact.objects.get(pk=contactid)
            cart = kwargs.get('cart', None)
            if not cart:
                cart = Cart.objects.from_request(request)
            self.order = get_or_create_order(request, cart, contact, self.cleaned_data)
            form_postsave.send(PaymentContactInfoForm, form=self)
            return contactid

        def clean(self):
            try:
                paymentmethod = self.cleaned_data['paymentmethod']
            except KeyError:
                self._errors['paymentmethod'] = forms.util.ErrorList([_('This field is required')])
                return self.cleaned_data
            required_fields = self.payment_required_fields.get(paymentmethod, [])
            msg = _('Selected payment method requires this field to be filled')
            for fld in required_fields:
                if not (self.cleaned_data.has_key(fld) and self.cleaned_data[fld]):
                    self._errors[fld] = forms.util.ErrorList([msg])
                elif fld == 'state':
                    self.enforce_state = True
                    try:
                        self._check_state(self.cleaned_data['state'], self.cleaned_data['country'])
                    except forms.ValidationError, e:
                        self._errors[fld] = e.messages
            super(PaymentContactInfoForm, self).clean()
            return self.cleaned_data
            
        def clean_discount(self):
            """ Check if discount exists and is valid. """
            if not config_value('PAYMENT', 'USE_DISCOUNTS'):
                return ''
            data = self.cleaned_data['discount']
            if data:
                try:
                    discount = Discount.objects.get(code=data, active=True)
                except Discount.DoesNotExist:
                    raise forms.ValidationError(_('Invalid discount code.'))
                    
                request = threadlocals.get_current_request()
                try:
                    contact = Contact.objects.from_request(request)
                except Contact.DoesNotExist:
                    contact = None

                valid, msg = discount.isValid(self.cart, contact=contact)

                if not valid:
                    raise forms.ValidationError(msg)
                # TODO: validate that it can work with these products
            return data
    

class SimplePayShipForm(forms.Form):
    shipping = forms.ChoiceField(widget=forms.RadioSelect(), required=False)

    def __init__(self, request, paymentmodule, *args, **kwargs):
        super(SimplePayShipForm, self).__init__(*args, **kwargs)
        
        try:
            order = Order.objects.from_request(request)
        except Order.DoesNotExist:
            order = None
        self.order = order
        self.orderpayment = None
        self.paymentmodule = paymentmodule
        
        try:
            self.tempCart = Cart.objects.from_request(request)
            if self.tempCart.numItems > 0:
                products = [item.product for item in self.tempCart.cartitem_set.all()]
            
        except Cart.DoesNotExist:
            self.tempCart = None
            
        try:
            self.tempContact = Contact.objects.from_request(request)
        except Contact.DoesNotExist:
            self.tempContact = None
            
        if kwargs.has_key('default_view_tax'):
            default_view_tax = kwargs['default_view_tax']
        else:
            default_view_tax = config_value_safe('TAX', 'TAX_SHIPPING', False)
            
        shipping_choices, shipping_dict = _get_shipping_choices(request, paymentmodule, self.tempCart, self.tempContact, default_view_tax=default_view_tax)
        
        
        cheapshipping = _get_cheapest_shipping(shipping_dict)
        self.cheapshipping = cheapshipping
        discount = None
        if order and order.discount_code:
            try:
                discount = Discount.objects.by_code(order.discount_code)
                if discount and discount.shipping == "FREECHEAP":
                    if cheapshipping:
                        shipping_choices = [opt for opt in shipping_choices if opt[0] == cheapshipping]
                        shipping_dict = {cheapshipping: shipping_dict[cheapshipping]}
            except Discount.DoesNotExist:
                pass
        
        # possibly hide the shipping        
        shiphide = config_value('SHIPPING','HIDING')
        
        if shiphide in ('YES', 'DESCRIPTION') and len(shipping_choices) == 1:
            self.fields['shipping'] = forms.CharField(max_length=30, initial=shipping_choices[0][0], 
                widget=forms.HiddenInput(attrs={'value' : shipping_choices[0][0]}))
            if shiphide == 'DESCRIPTION':
                self.shipping_hidden = False
                self.shipping_description = shipping_choices[0][1]
            else:
                self.shipping_hidden = True
                self.shipping_description = ""
        elif len(shipping_choices) == 0:
            self.shipping_hidden = True
        else:
            self.fields['shipping'].choices = shipping_choices
            if config_value('SHIPPING','SELECT_CHEAPEST'):
                if cheapshipping is not None:
                    self.fields['shipping'].initial = cheapshipping
            self.shipping_hidden = False

        self.shipping_dict = shipping_dict
        
        form_init.send(SimplePayShipForm, form=self)
        
    def clean_shipping(self):
        shipping = self.cleaned_data['shipping']
        if not shipping and self.tempCart.is_shippable:
            raise forms.ValidationError(_('This field is required.'))
        return shipping

    def is_needed(self):
        """Check to see if this form is even needed
        it is *not* needed if:
        - we have an order
        - the order balance is zero
        - No shipping needs to be selected
        """

        needed = True
        if self.order and self.tempContact and self.tempCart:
            order = self.order
            if order.is_shippable and len(self.shipping_dict) == 1:
                update_shipping(order, self.shipping_dict.keys()[0], self.tempContact, self.tempCart)
            order.recalculate_total(save=False)
            needed = not order.paid_in_full
            if not needed:
                log.debug('%s can skip the payment step - no info needed', order)
                
        return needed

    def save(self, request, cart, contact, payment_module, data=None):
        form_presave.send(SimplePayShipForm, form=self)
        if data is None:
            data = self.cleaned_data
        self.order = get_or_create_order(request, cart, contact, data)
        if payment_module:
            processor_module = payment_module.MODULE.load_module('processor')
            processor = processor_module.PaymentProcessor(payment_module)
            self.orderpayment = processor.create_pending_payment(order=self.order)
        else:
            self.orderpayment = None
        form_postsave.send(SimplePayShipForm, form=self)


class CreditPayShipForm(SimplePayShipForm):
    credit_type = forms.ChoiceField()
    credit_number = forms.CharField(max_length=20, widget=forms.TextInput(attrs={'autocomplete':'off'}))
    month_expires = forms.ChoiceField(choices=MONTHS)
    year_expires = forms.ChoiceField()
    ccv = forms.CharField(max_length=4, label='Sec code', widget=forms.TextInput(attrs={'autocomplete':'off'}))

    def __init__(self, request, paymentmodule, *args, **kwargs):
        creditchoices = paymentmodule.CREDITCHOICES.choice_values
        super(CreditPayShipForm, self).__init__(request, paymentmodule, *args, **kwargs)

        self.cc = None

        self.fields['credit_type'].choices = creditchoices
        
        num_years = config_value('PAYMENT', 'CC_NUM_YEARS')
        year_now = datetime.date.today().year
        self.fields['year_expires'].choices = [(year, year) for year in range(year_now, year_now+num_years+1)]

        self.tempCart = Cart.objects.from_request(request)
        
        initial = kwargs.get('initial', None)
        if initial:
            if initial.get('credit_number', None):
                self.fields['credit_number'].widget = forms.PasswordInput()
            if initial.get('ccv', None):
                self.fields['ccv'].widget = forms.PasswordInput()
            
        try:
            self.tempContact = Contact.objects.from_request(request)
        except Contact.DoesNotExist:
            self.tempContact = None

    def clean(self):            
        super(CreditPayShipForm, self).clean()
        data = self.cleaned_data
        if not self.is_valid():
            log.debug('form not valid, no early auth')
            return data
        early = config_value('PAYMENT', 'AUTH_EARLY')
        
        if early:
            processor_module = self.paymentmodule.MODULE.load_module('processor')
            processor = processor_module.PaymentProcessor(self.paymentmodule)
            if processor.can_authorize():
                log.debug('Processing early capture/release for: %s', self.order)
                processor_module = self.paymentmodule.MODULE.load_module('processor')
                processor = processor_module.PaymentProcessor(self.paymentmodule)                
                if self.order:
                    # we have to make a payment object and save the credit card data to
                    # make an auth/release.
                    orderpayment = processor.create_pending_payment(order=self.order, 
                        amount=Decimal('0.01'))
                    op = orderpayment.capture

                    cc = CreditCardDetail(orderpayment=op,
                        expire_month=data['month_expires'],
                        expire_year=data['year_expires'],
                        credit_type=data['credit_type'])

                    cc.storeCC(data['credit_number'])
                    cc.save()

                    # set ccv into cache
                    cc.ccv = data['ccv']
                    self.cc = cc
                    results = processor.authorize_and_release(order=self.order)
                if not results.success:
                    log.debug('Payment module error: %s', results)
                    raise forms.ValidationError(results.message)
                else:
                    log.debug('Payment module capture/release success for %s', self.order)
            else:
                log.debug('Payment module %s cannot do credit authorizations, ignoring AUTH_EARLY setting.',
                    self.paymentmodule.MODULE.KEY.value)
        return data
        

    def clean_credit_number(self):
        """ Check if credit card is valid. """
        data = self.cleaned_data
        credit_number = data['credit_number']
        card = CreditCard(credit_number, data['credit_type'])
        results, msg = card.verifyCardTypeandNumber()
        if not results:
            raise forms.ValidationError(msg)

        return credit_number

    def clean_month_expires(self):
        return int(self.cleaned_data['month_expires'])

    def clean_year_expires(self):
        """ Check if credit card has expired. """
        month = self.cleaned_data['month_expires']
        year = int(self.cleaned_data['year_expires'])
        max_day = calendar.monthrange(year, month)[1]
        if datetime.date.today() > datetime.date(year=year, month=month, day=max_day):
            raise forms.ValidationError(_('Your card has expired.'))
        return year

    def clean_ccv(self):
        """ Validate a proper CCV is entered. Remember it can have a leading 0 so don't convert to int and return it"""
        try:
            check = int(self.cleaned_data['ccv'])
            return self.cleaned_data['ccv'].strip()
        except ValueError:
            raise forms.ValidationError(_('Invalid ccv.'))
            
    def save(self, request, cart, contact, payment_module, data=None):
        """Save the order and the credit card information for this orderpayment"""
        form_presave.send(CreditPayShipForm, form=self)
        if data is None:
            data = self.cleaned_data
        assert(data)
        super(CreditPayShipForm, self).save(request, cart, contact, payment_module, data=data)
        
        if self.orderpayment:
            op = self.orderpayment.capture

            cc = CreditCardDetail(orderpayment=op,
                expire_month=data['month_expires'],
                expire_year=data['year_expires'],
                credit_type=data['credit_type'])

            cc.storeCC(data['credit_number'])
            cc.save()

            # set ccv into cache
            cc.ccv = data['ccv']
            self.cc = cc
        form_postsave.send(CreditPayShipForm, form=self)



from notmm.utils.urlmap import RegexURLMap

urlpatterns = RegexURLMap()

#urlpatterns.set_default_prefix('/foo/')

urlpatterns.add_routes('sandbox.views.decorators', 
        ### decorators tests
        ### with_schevo_database test
        (r'^test_with_schevo_database$', 'test_schevo_database', {
            'template_name' : 'sandbox/test.mako'
        }),
)

urlpatterns.add_routes('sandbox.handlers',
        (r'^moved/$', 'handle302', {
            'location' : '/index.html'
        }),
        (r'^croak$', 'handle500')
)
#urlpatterns.commit()




# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import io
from gettext import dgettext
import xml.dom
from xml.sax.saxutils import escape
import gobject
import gtk

from gazpacho import util
from gazpacho.choice import enum_to_string, flags_to_string
from gazpacho.placeholder import Placeholder
from gazpacho.gadget import Gadget
from gazpacho.properties import prop_registry, ObjectType

def write_xml(file, xml_node, indent=0, indent_increase=4):
    if xml_node.nodeType == xml_node.TEXT_NODE:
        file.write(xml_node.data)
        return
    elif xml_node.nodeType == xml_node.CDATA_SECTION_NODE:
        file.write('<![CDATA[%s]]>' % xml_node.data)
        return

    file.write(' '*indent)

    file.write('<%s' % xml_node.tagName)
    if len(xml_node.attributes) > 0:
        attr_string = ' '.join(['%s="%s"' % (n, v)
                                    for n, v in list(xml_node.attributes.items())])
        file.write(' ' + attr_string)

    children = [a for a in xml_node.childNodes
                    if a.nodeType != a.ATTRIBUTE_NODE]
    if children:
        has_text_child = False
        for child in children:
            if child.nodeType in (child.TEXT_NODE,
                                  child.CDATA_SECTION_NODE):
                has_text_child = True
                break

        if has_text_child:
            file.write('>')
        else:
            file.write('>\n')
        for child in children:
            write_xml(file, child, indent+indent_increase, indent_increase)

        if not has_text_child:
            file.write(' '*indent)
        file.write('</%s>\n' % xml_node.tagName)
    else:
        file.write('/>\n')

class XMLWriter:
    """
    @ivar _doc:
    @ivar _project:
    @ivar _version:
    @ivar _skip_external_references: If True we clear all object
       references that are not part of the current serialization
    @ivar _all_widgets: all widgets in the current serialization
    """
    def __init__(self, document=None, project=None):
        if not document:
            dom = xml.dom.getDOMImplementation()
            document = dom.createDocument(None, None, None)
        self._doc = document
        self._project = project
        self._version = "libglade"

        self._skip_external_references = False
        self._all_widgets = []

    def write(self, path, widgets, sizegroups, models, uim, domain, version):
        self._version = version
        root = self._write_root(widgets, domain)
        self._write_sizegroups(root, sizegroups)
        for model in models:
            self._write_model(root, model)
        self._write_widgets(root, widgets, uim)
        self.write_node(path, root)

    def _write_header(self, f):
        f.write('<?xml version="1.0" standalone="no"?> <!--*- mode: xml -*-->\n')
        doctype = None
        if self._version == 'libglade':
            doctype = "http://gazpacho.sicem.biz/gazpacho-0.1.dtd"

        if doctype:
            f.write('<!DOCTYPE glade-interface SYSTEM "%s">\n' % doctype)

    # FIXME: Should this really be exported
    def write_node(self, path, root):
        f = file(path, 'w')
        self._write_header(f)
        write_xml(f, root)
        f.close()

    def _create_root(self):
        if self._version == "libglade":
            tag = 'glade-interface'
        else:
            tag = 'interface'

        element = self._doc.createElement(tag)
        return self._doc.appendChild(element)

    def serialize_node(self, gadget):
        root = self._write_root()

        # save the UI Manager if needed
        widgets = gadget.project.uim.get_ui_widgets(gadget.widget)
        if widgets:
            uim_node = self._write_ui_manager(gadget.project.uim, widgets)
            if uim_node:
                root.appendChild(uim_node)

        # save the Model for TreeViews
        if isinstance(gadget.widget, gtk.TreeView):
            model = gadget.adaptor.get_model(gadget)
            self._write_model(root, model)

        root.appendChild(self._write_widget(gadget))

        return self._doc.documentElement

    def serialize(self, gadget, skip_external_references=False):

        if skip_external_references:
            self._skip_external_references = True
            self._all_widgets = get_all_widgets(gadget.widget)

        fp = io.StringIO()
        node = self.serialize_node(gadget)
        write_xml(fp, node)
        fp.seek(0)

        # reset context dependent states
        self._skip_external_references = False
        self._all_widgets = []

        return fp.read()

    def serialize_widgets(self, widgets, sizegroups, models, uim):
        fp = io.StringIO()
        root = self._write_root(widgets)
        self._write_sizegroups(root, sizegroups)
        for model in models:
            self._write_model(root, model)
        self._write_widgets(root, widgets, uim)
        write_xml(fp, root)
        fp.seek(0)

        return fp.read()

    def _get_requirements(self, widgets):
        # check what modules are the gadgets using

        modules = []
        for widget in widgets:
            gadget = Gadget.from_widget(widget)
            if gadget:
                libglade_module = gadget.adaptor.library.library_name
                if libglade_module and libglade_module not in modules:
                    modules.append(libglade_module)

        # the base module is not a requirement
        if 'base' in modules:
            modules.remove('base')

        return modules

    def _write_root(self, widgets=None, domain=None):
        root = self._create_root()

        if domain:
            root.setAttribute('domain', domain)

        # Disable require for now, see bug #344672
        if 0: #self._version == "libglade" and widgets:
            for module in self._get_requirements(widgets):
                n = self._doc.createElement('requires')
                n.setAttribute('lib', module)
                root.appendChild(n)

        return root

    def _create_object(self, class_name, name):
        if self._version == "gtkbuilder":
            tag = 'object'
        else:
            tag = 'widget'
        node = self._doc.createElement(tag)
        node.setAttribute('class', class_name)
        node.setAttribute('id', name)
        return node

    def _write_sizegroups(self, root, sizegroups):
        for sizegroup in sizegroups:
            node = self._create_object('GtkSizeGroup', sizegroup.name)
            root.appendChild(node)
            prop_node = self._doc.createElement('property')
            node.appendChild(prop_node)
            prop_node.setAttribute('name', 'mode')
            text = self._doc.createTextNode(enum_to_string(
                sizegroup.mode, enum=gtk.SizeGroupMode))
            prop_node.appendChild(text)

            # This needs to be cleaned up, ideally moved to sizegroup.py
            if self._version == 'gtkbuilder':
                widgets = self._doc.createElement('widgets')
                node.appendChild(widgets)
                for gadget in sizegroup.get_gadgets():
                    widget = self._doc.createElement('widget')
                    widget.setAttribute('name', gadget.name)
                    widgets.appendChild(widget)

    def _write_ui_manager(self, uim, widgets):
        """Create a XML node with the information of this UIManager

        Only the information about the widgets of this UIManager is
        actually saved.
        """

        action_groups = uim.get_action_groups()
        if not action_groups:
            return

        uim_node = self._create_object('GtkUIManager', uim.get_name())

        # first save action groups
        # XXX don't save all the action groups, only those ones
        # that are in use
        for action_group in action_groups:
            child_node = self._doc.createElement('child')
            uim_node.appendChild(child_node)
            child_node.appendChild(self._write_action_group(action_group))

        # Append uimanager
        gadgets = []
        for widget in widgets:
            if not isinstance(widget, (gtk.Toolbar, gtk.MenuBar)):
                continue
            gadget = Gadget.from_widget(widget)
            if gadget is None:
                raise AssertionError("There is no gadget for %s" % widget)
            gadgets.append(gadget)

        nodes = uim.save_ui_definitions(self._doc, gadgets, self._version)
        for ui_node in nodes:
            uim_node.appendChild(ui_node)

        return uim_node

    def _write_action_group(self, action_group):
        node = self._create_object('GtkActionGroup', action_group.name)

        # Sort all children, to keep the file format stabler
        action_names = action_group.get_action_names()
        action_names.sort()

        for action_name in action_names:
            child_node = self._doc.createElement('child')
            node.appendChild(child_node)
            action = action_group.get_action(action_name)
            child_node.appendChild(self._write_action(action))

        return node

    def _write_action(self, action):
        # XXX: Use a gadget and property wrappers

        action_node = self._create_object('GtkAction', action.name)

        default_label = None
        default_key = 0
        default_mask = None
        domain = None
        if action.stock_id:
            stock_info = gtk.stock_lookup(action.stock_id)
            if stock_info:
                (_, default_label, default_mask,
                 default_key, domain) = stock_info

        node = self._write_property('name', action.name)
        action_node.appendChild(node)

        # default_label is translated, so compare against the
        # untranslated version sent through dgettext()
        if default_label != dgettext(domain, action.label):
            node = self._write_property('label', action.label, True)
            action_node.appendChild(node)

        if action.label != action.short_label:
            node = self._write_property('short_label', action.short_label,
                                        True)
            action_node.appendChild(node)

        if action.is_important:
            node = self._write_property('is_important', 'True')
            action_node.appendChild(node)

        if action.tooltip:
            node = self._write_property('tooltip', action.tooltip, True)
            action_node.appendChild(node)

        if action.stock_id:
            node = self._write_property('stock_id', action.stock_id)
            action_node.appendChild(node)

        if action.callback:
            signal_node = self._write_signal('activate', action.callback)
            action_node.appendChild(signal_node)

        if action.accelerator and self._version != 'gtkbuilder':
            key, mask = gtk.accelerator_parse(action.accelerator)
            if key != default_key or mask != default_mask:
                node = self._write_property('accelerator',
                                            action.accelerator)
                action_node.appendChild(node)
        return action_node

    def _write_model(self, root, model):
        # <widget>
        #   <columns>
        #     <column type="...">
        #   </columns>
        # <widget/>

        model_node = self._create_object(model.get_type_name(),
                                         model.get_name())
        root.appendChild(model_node)

        # first we save the types
        column_type_names = model.get_column_type_names()
        if column_type_names:
            columns_node = self._doc.createElement('columns')
            model_node.appendChild(columns_node)
            for type_name in column_type_names:
                column_node = self._doc.createElement('column')
                column_node.setAttribute('type', type_name)
                columns_node.appendChild(column_node)

        # then the data
        store = model.get_model()
        if len(store):
            data_node = self._write_list_store_data(store,
                                                    len(column_type_names))
            model_node.appendChild(data_node)

    def _write_list_store_data(self, store, n_columns):
        # <data>
        #   <row>
        #     <col id="..">...</col>
        #   </row>
        # </data>

        data_node = self._doc.createElement('data')
        for row in store:
            row_node = self._doc.createElement('row')
            data_node.appendChild(row_node)

            for i in range(n_columns):
                col_node = self._doc.createElement('col')
                row_node.appendChild(col_node)
                col_node.setAttribute('id', str(i))

                data = row[i]
                if data is None:
                    text = ''
                else:
                    text = escape(str(row[i]))
                node = self._doc.createTextNode(text)
                col_node.appendChild(node)

        return data_node

    def _write_widgets(self, node, widgets, uim):
        ui_node = self._write_ui_manager(uim, widgets)
        if ui_node:
            node.appendChild(ui_node)

        # Append toplevel widgets. Each widget then takes care of
        # appending its children
        for widget in widgets:
            gadget = Gadget.from_widget(widget)
            if (gadget is None or
                not gadget.is_toplevel()):
                continue

            wnode = self._write_widget(gadget)
            node.appendChild(wnode)

        return node

    def _write_widget(self, gadget):
        """Serializes this gadget into a XML node and returns this node"""

        gadget.maintain_gtk_properties = True

        gadget.adaptor.save(gadget.project.context, gadget)

        # otherwise use the default saver
        node = self._write_basic_information(gadget)

        self._write_properties(node, gadget, child=False)
        for signals in gadget.get_all_signal_handlers():
            for signal in signals:
                signal_node = self._write_signal(signal.name, signal.handler,
                                                 signal.after, signal.object)
                node.appendChild(signal_node)

        # Children

        # We're not writing children when we have a constructor set
        if gadget.constructor:
            return node

        widget = gadget.widget
        children = gadget.get_children()
        if isinstance(widget, gtk.Table):
            table = widget
            def table_sort(a, b):
                res =  cmp(table.child_get_property(a, 'left-attach'),
                           table.child_get_property(b, 'left-attach'))
                if res == 0:
                    res = cmp(table.child_get_property(a, 'top-attach'),
                              table.child_get_property(b, 'top-attach'))
                return res
            children.sort(table_sort)

        # Boxes doesn't need to be sorted, they're already in the right order
        for child_widget in children:
            if isinstance(child_widget, Placeholder):
                child_node = self._write_placeholder(child_widget)
            elif isinstance(child_widget, gtk.CellRenderer):
                child_node = self._write_cell_renderer(child_widget,
                                                       gadget.widget)
            else:
                # if there is no gadget for this child
                # we don't save it. If your children are not being
                # saved you should create a gadget for them
                # in your Adaptor
                child_gadget = Gadget.from_widget(child_widget)
                if child_gadget is None:
                    continue
                child_node = self._write_child(child_gadget)

            node.appendChild(child_node)

        gadget.maintain_gtk_properties = False

        return node

    def _write_properties(self, parent_node, gadget, child):
        """
        @parent_node: the node of the parent, could be a widget node or a
           child node when setting packing properties
        @gadget:
        @child: boolean, if true write child properties, if false write normal
        """

        # <property name>...</property>
        # <property name>...</property>
        # ...
        properties = prop_registry.get_writable_properties(gadget.widget,
                                                           child)
        properties.sort(lambda a, b: cmp(a.name, b.name))

        for prop_type in properties:

            if prop_type.child:
                prop = gadget.get_child_prop(prop_type.name)
            else:
                prop = gadget.get_prop(prop_type.name)

            if isinstance(prop, ObjectType):
                if self._skip_external_references:
                    if prop.value not in self._all_widgets:
                        continue

            value = prop.save()
            if value == None:
                continue
            name = prop.name.replace('-', '_')
            translatable = False
            context = False
            comment = None
            if prop.is_translatable():
                translatable = True
                if prop.has_i18n_context:
                    context = True
                if prop.i18n_comment:
                    comment = prop.i18n_comment
            property_node = self._write_property(name, value, translatable,
                                                 context, comment)
            parent_node.appendChild(property_node)

    def _write_signal(self, name, handler, after=False, object=None):
        # <signal name="..." handler="..." after="..." object="..."/>
        signal_node = self._doc.createElement('signal')
        signal_node.setAttribute('name', name)
        signal_node.setAttribute('handler', handler)
        if after:
            signal_node.setAttribute('after', 'True')
        if object:
            signal_node.setAttribute('object', object)
        return signal_node

    def _write_basic_information(self, gadget):
        # <widget class="..." id=" constructor="">
        assert gadget.adaptor.type_name
        assert gadget.name, 'gadget %r is nameless' % gadget.widget

        # If name is set write it to the file
        if gadget.adaptor.name:
            type_name = gadget.adaptor.name
        else:
            type_name = gadget.adaptor.type_name

        node = self._create_object(type_name, gadget.name)
        if gadget.constructor:
            constructor = gadget.constructor
            node.setAttribute('constructor', constructor)
        return node

    def _write_child(self, child_gadget):
        # <child>
        #   <child internal-name="foo">
        # </child>

        child_node = self._doc.createElement('child')

        if child_gadget.internal_name is not None:
            child_node.setAttribute('internal-child',
                                    child_gadget.internal_name)

        child = self._write_widget(child_gadget)
        child_node.appendChild(child)

        # Append the packing properties
        packing_node = self._doc.createElement('packing')
        self._write_properties(packing_node, child_gadget, child=True)

        if packing_node.childNodes:
            child_node.appendChild(packing_node)

        return child_node

    def _write_placeholder(self, widget):
        # <child>
        #   <placeholder>
        # </child>

        child_node = self._doc.createElement('child')
        placeholder_node = self._doc.createElement('placeholder')
        child_node.appendChild(placeholder_node)

        # we need to write the packing properties of the placeholder.
        # otherwise the container gets confused when loading its
        # children
        node = self._write_placeholder_properties(widget)
        if node:
            child_node.appendChild(node)

        return child_node

    def _write_placeholder_properties(self, placeholder):
        # <placholder/>
        # <packing>
        #    <property>...</property>
        #    ....
        # </packing>

        parent = placeholder.get_parent()
        pspecs = gtk.container_class_list_child_properties(parent)
        if not pspecs:
            return
        packing_node = self._doc.createElement('packing')
        for pspec in pspecs:
            value = parent.child_get_property(placeholder, pspec.name)
            if value == pspec.default_value:
                continue
            prop_name = pspec.name.replace('-', '_')

            if gobject.type_is_a(pspec.value_type, gobject.TYPE_ENUM):
                v = enum_to_string(value, pspec)
            elif gobject.type_is_a(pspec.value_type, gobject.TYPE_FLAGS):
                v = flags_to_string(value, pspec)
            else:
                v = escape(str(value))

            prop_node = self._write_property(prop_name, v)
            packing_node.appendChild(prop_node)

        return packing_node

    def _write_property(self, name, value, translatable=False,
                        context=False, comment=False):
        # <property name="..."
        #           context="yes|no"
        #           translatable="yes|no">...</property>

        node = self._doc.createElement('property')

        # We should change each '-' by '_' on the name of the property
        # put the name="..." part on the <property ...> tag
        node.setAttribute('name', name)

        # Only write context and comment if translatable is
        # enabled, to mimic glade-2
        if translatable:
            node.setAttribute('translatable', 'yes')
            if context and self._version != 'gtkbuilder':
                node.setAttribute('context', 'yes')
            if comment:
                node.setAttribute('comment', comment)

        text = self._doc.createTextNode(escape(value))
        node.appendChild(text)
        return node

    def _write_cell_renderer(self, cell_renderer, column):
        child_node = self._doc.createElement('child')
        name = column.get_name_for_cell_renderer(cell_renderer)
        cell_node = self._create_object(gobject.type_name(cell_renderer), name)
        child_node.appendChild(cell_node)

        attributes_node = self._doc.createElement('attributes')
        child_node.appendChild(attributes_node)

        attributes = column.get_attributes(cell_renderer)
        for key, value in list(attributes.items()):
            attr_node = self._doc.createElement('attribute')
            attr_node.setAttribute('name', key)
            text = self._doc.createTextNode(escape(str(value)))
            attr_node.appendChild(text)
            attributes_node.appendChild(attr_node)

        return child_node


def get_all_widgets(widget):
    if isinstance(widget, Placeholder):
        return []

    widgets = [widget]
    if isinstance(widget, gtk.Container):
        for child in util.get_all_children(widget):
            widgets += get_all_widgets(child)
    return widgets



from django.conf import settings
from django.contrib.comments.models import Comment
from django.contrib.sites.models import Site
from django.db.models import Avg
from django.utils.translation import ugettext_lazy as _
from satchmo_ext.productratings.models import ProductRating
import logging
import operator

log = logging.getLogger('product.comments.utils')

def average(ratings):
    """ Average a list of numbers, return None if it fails """
    if ratings:
        ratings = filter(lambda x: x is not None, ratings)
    if ratings:
        total = reduce(operator.add, ratings)
        if total != None:
            return float(total)/len(ratings)
    return None

def get_product_rating(product, site=None):
    """Get the average product rating"""
    if site is None:
        site = Site.objects.get_current()
    
    site = site.id
    manager = Comment.objects
    comment_pks = manager.filter(object_pk__exact=str(product.id),
                               content_type__app_label__exact='product',
                               content_type__model__exact='product',
                               site__id__exact=site,
                               is_public__exact=True).values_list('pk', flat=True)
    rating = ProductRating.objects.filter(comment__in=comment_pks
        ).aggregate(average=Avg('rating'))['average']
    log.debug("Rating: %s", rating)
    return rating

def get_product_rating_string(product, site=None):
    """Get the average product rating as a string, for use in templates"""
    rating = get_product_rating(product, site=site)
    
    if rating is not None:
        rating = "%0.1f" % rating
        if rating.endswith('.0'):
            rating = rating[0]
        rating = rating + "/5"
    else:
        rating = _('Not Rated')
        
    return rating
    



from django.shortcuts import render_to_response
from django.template import RequestContext

def example(request):
    ctx = RequestContext(request, {})
    return render_to_response('localsite/example.html', context_instance=ctx)



from decimal import Decimal
from django import template
from django.conf import settings
from django.core import urlresolvers
from django.template import Context, Template
from django.utils.translation import get_language, ugettext_lazy as _
from livesettings import config_value
from product.utils import calc_discounted_by_percentage, find_best_auto_discount
from satchmo_utils.templatetags import get_filter_args
from tax.templatetags import satchmo_tax

register = template.Library()

def sale_price(product):
    """Returns the sale price, including tax if that is the default."""
    if config_value('TAX', 'DEFAULT_VIEW_TAX'):
        return taxed_sale_price(product)
    else:
        return untaxed_sale_price(product)

register.filter('sale_price', sale_price)

def untaxed_sale_price(product):
    """Returns the product unit price with the best auto discount applied."""
    discount = find_best_auto_discount(product)
    price = product.unit_price
        
    if discount and discount.valid_for_product(product):
        price = calc_discounted_by_percentage(price, discount.percentage)
    
    return price

register.filter('untaxed_sale_price', untaxed_sale_price)

def taxed_sale_price(product):
    """Returns the product unit price with the best auto discount applied and taxes included."""
    taxer = satchmo_tax._get_taxprocessor()
    price = untaxed_sale_price(product)
    price = price + taxer.by_price(product.taxClass, price)
    return price

register.filter('taxed_sale_price', taxed_sale_price)

def discount_cart_total(cart, discount):
    """Returns the discounted total for this cart, with tax if that is the default."""
    if config_value('TAX', 'DEFAULT_VIEW_TAX'):
        return taxed_discount_cart_total(cart, discount)
    else:
        return untaxed_discount_cart_total(cart, discount)

register.filter('discount_cart_total', discount_cart_total)

def untaxed_discount_cart_total(cart, discount):
    """Returns the discounted total for this cart"""
    total = Decimal('0.00')
    
    for item in cart:
        total += untaxed_discount_line_total(item, discount)
    return total
        
register.filter('untaxed_discount_cart_total', untaxed_discount_cart_total)

def taxed_discount_cart_total(cart, discount):
    """Returns the discounted total for this cart with taxes included"""
    total = Decimal('0.00')
    
    for item in cart:
        total += taxed_discount_line_total(item, discount)
        
    return total
        
register.filter('taxed_discount_cart_total', taxed_discount_cart_total)

def discount_line_total(cartitem, discount):
    """Returns the discounted line total for this cart item, including tax if that is the default."""
    if config_value('TAX', 'DEFAULT_VIEW_TAX'):
        return taxed_discount_line_total(cartitem, discount)
    else:
        return untaxed_discount_line_total(cartitem, discount)

    
register.filter('discount_line_total', discount_line_total)

def untaxed_discount_line_total(cartitem, discount):
    """Returns the discounted line total for this cart item"""
    price = cartitem.line_total
    if discount and discount.valid_for_product(cartitem.product):
        price = calc_discounted_by_percentage(price, discount.percentage)
    
    return price
        
register.filter('untaxed_discount_line_total', untaxed_discount_line_total)

def taxed_discount_line_total(cartitem, discount):
    """Returns the discounted line total for this cart item with taxes included."""
    price = untaxed_discount_line_total(cartitem, discount)
    taxer = satchmo_tax._get_taxprocessor()
    price = price + taxer.by_price(cartitem.product.taxClass, price)
    
    return price
        
register.filter('taxed_discount_line_total', taxed_discount_line_total)

def discount_price(product, discount):
    """Returns the product price with the discount applied, including tax if that is the default.
    
    Ex: product|discount_price:sale
    """
    if config_value('TAX', 'DEFAULT_VIEW_TAX'):
        return taxed_discount_price(product, discount)
    else:
        return untaxed_discount_price(product, discount)


register.filter('discount_price', discount_price)
   
def untaxed_discount_price(product, discount):
    """Returns the product price with the discount applied.
    
    Ex: product|discount_price:sale
    """
    up = product.unit_price
    if discount and discount.valid_for_product(product):
        pcnt = calc_discounted_by_percentage(up, discount.percentage)
        return pcnt
    else:
        return up

register.filter('untaxed_discount_price', untaxed_discount_price)

def taxed_discount_price(product, discount):
    """Returns the product price with the discount applied, and taxes included.
    
    Ex: product|discount_price:sale
    """
    price = untaxed_discount_price(product, discount)
    taxer = satchmo_tax._get_taxprocessor()
    return price + taxer.by_price(product.taxClass, price)

register.filter('taxed_discount_price', taxed_discount_price)

def discount_ratio(discount):
    """Returns the discount as a ratio, making sure that the percent is under 1"""
    pcnt = discount.percentage
    if pcnt > 1:
        pcnt = pcnt/100
    
    return 1-pcnt

register.filter('discount_ratio', discount_ratio)

def discount_saved(product, discount):
    """Returns the amount saved by the discount, including tax if that is the default."""
    if config_value('TAX', 'DEFAULT_VIEW_TAX'):
        return taxed_discount_saved(product, discount)
    else:
        return untaxed_discount_saved(product, discount)

register.filter('discount_saved', discount_saved)


def untaxed_discount_saved(product, discount):
    """Returns the amount saved by the discount"""
    
    if discount and discount.valid_for_product(product):
        price = product.unit_price
        discounted = untaxed_discount_price(product, discount)
        saved = price - discounted
        cents = Decimal("0.01")
        return saved.quantize(cents)
    else:
        return Decimal('0.00')

register.filter('untaxed_discount_saved', untaxed_discount_saved)

def taxed_discount_saved(product, discount):
    """Returns the amount saved by the discount, after applying taxes."""
    
    if discount and discount.valid_for_product(product):
        price = product.unit_price
        discounted = taxed_discount_price(product, discount)
        saved = price - discounted
        cents = Decimal("0.01")
        return saved.quantize(cents)
    else:
        return Decimal('0.00')

register.filter('taxed_discount_saved', taxed_discount_saved)



# Copyright (C) 2005 by Async Open Source
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gazpacho.properties import prop_registry, CustomProperty, \
     ProxiedProperty, AdjustmentType, FloatType, IntType

class SpinAdjustmentType(AdjustmentType):
    def __init__(self, gobj):
        super(AdjustmentType, self).__init__(gobj)
        # Initialize us to the value fetch from the loader
        self._value = gobj.widget.get_adjustment()

prop_registry.override_simple('GtkSpinButton::adjustment', SpinAdjustmentType)

# Value is never saved, because it's already saved as the first item in the
# adjustment. Infact get_property in gtkspinbutton.c just returns
# spinbutton->adjustment->value, so it makes sense to ignore it.
class Value(CustomProperty):
    def save(self):
        pass
prop_registry.override_simple('GtkSpinButton::value', Value)

class AdjustmentProxy(ProxiedProperty):
    def get_target(self):
        return self.object.get_adjustment()

    def save(self):
        # Not saving, they're saved in GtkSpinButton::adjustment instead
        return

class Lower(AdjustmentProxy, FloatType):
    label = "Min"
    default = 0
    tooltip = "The minimum value the spinbutton can have"
    target_name = 'lower'
prop_registry.override_property('GtkSpinButton::lower', Lower)

class Upper(AdjustmentProxy, FloatType):
    label = "Max"
    default = 100
    tooltip = "The maximum value the spinbutton can have"
    target_name = 'upper'
prop_registry.override_property('GtkSpinButton::upper', Upper)

class Step(AdjustmentProxy, IntType):
    label = "Step increment"
    default = 1
    tooltip = "Increment applied for each left mousebutton press"
    target_name = 'step-increment'
prop_registry.override_property('GtkSpinButton::step-increment', Step)

class Page(AdjustmentProxy, IntType):
    label = "Page increment"
    default = 10
    tooltip = "Increment applied for each middle mousebutton press"
    target_name = 'page-increment'
prop_registry.override_property('GtkSpinButton::page-increment', Page)



#!/usr/bin/env python
"""This module allows communication with Schevo named
databases using the Durus Extended Server.

You'll need the ``xdserver`` package and the ``schevodurus``
packages. In addition, a minimal version of Schevo 3.1 is
also required.

Test case is in ``dbapi/schevo/test_xdserver.py``.
"""

# Copyright (c) 2010 Etienne Robillard <erob@gthcfoundation.org>
# All rights reserved.

#from schevodurus.backend import DurusBackend as Connection
#from durus.connection import Connection
#from durus.btree import BTree
from schevo.backends.durus39 import DurusBackend as Connection
from schevo.database import format_dbclass, equivalent
from schevo import icon
import schevo.mt
from test_support import unittest
from notmm.dbapi.orm import XdserverProxy

class XdserverProxyTestCase(unittest.TestCase):

    connected = False

    def setUp(self):
        self.dbname = 'moviereviews'

    def tearDown(self):
        pass

    def test_equivalent(self):
        db = XdserverProxy(self.dbname)
        db2 = XdserverProxy(self.dbname)
        self.assertEqual(equivalent(db, db2), True)
        #db.close()
        #db2.close()



# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gtk

from gazpacho.placeholder import Placeholder
from gazpacho.properties import prop_registry, CustomProperty, IntType
from gazpacho.gadget import Gadget
from gazpacho.widgets.base.base import ContainerAdaptor

class NotebookAdaptor(ContainerAdaptor):

    def _selection_changed_cb(self, selection, notebook):
        if not selection:
            return

        # Use the first widget that is embedded in the notebook
        for widget in selection:
            if widget.is_ancestor(notebook):
                break
        else:
            # No widget were embedded in us
            return

        # Now get the toplevel child embedded in the page
        while widget.get_parent() != notebook:
            widget = widget.get_parent()

        # Finally, if the page of that widget is different from the
        # currently selected one, switch to it
        widget_page = notebook.page_num(widget)
        if widget_page != notebook.get_current_page():
            notebook.set_current_page(widget_page)

    def _add_signal_handler(self, project, widget):
        project.selection.connect('selection-changed',
                                  self._selection_changed_cb, widget)

    def create(self, context, interactive=True):
        widget = super(NotebookAdaptor, self).create(context, interactive)
        self._add_signal_handler(context.get_project(), widget)
        return widget

    def load(self, context, widget):
        gadget = super(NotebookAdaptor, self).load(context, widget)
        self._add_signal_handler(context.get_project(), widget)
        return gadget

    def replace_child(self, context, current, new, container):
        page_num = container.page_num(current)
        if page_num == -1:
            return

        label = container.get_tab_label(current)

        container.remove_page(page_num)
        container.insert_page(new, label, page_num)

        new.show()
        container.set_current_page(page_num)


# GtkNotebook

class NPagesProp(CustomProperty, IntType):
    minimum = 1
    maximum = 100
    step_increment = 1
    page_increment = 1
    climb_rate = 1
    label = "Number of pages"
    default = 3
    persistent = False

    def load(self):
        self._initial = self.default

        # Don't set default if object has childs already
        if not self.get():
            self.set(self.default)

    def get(self):
        return self.object.get_n_pages()

    def set(self, value):
        old_size = self.object.get_n_pages()
        new_size = value
        if new_size == old_size:
            return

        if new_size > old_size:
            project = self._project
            # The notebook has grown. Add pages
            while new_size > old_size:
                label = gtk.Label()
                project.set_new_widget_name(label)

                #print load_gadget_from_widget(label, project)
                no = self.object.append_page(
                    Placeholder(), label)
                label.set_text('Page %d' % (no + 1))
                #project.add_widget(label)
                old_size += 1
        else:
            # The notebook has shrunk. Remove pages

            # Thing to remember is that GtkNotebook starts the
            # page numbers from 0, not 1 (C-style). So we need to do
            # old_size-1, where we're referring to "nth" widget.
            while old_size > new_size:
                child_widget = self.object.get_nth_page(old_size - 1)
                child_gadget = Gadget.from_widget(child_widget)
                # If we got it, and it's not a placeholder, remove it
                # from the project
                if child_gadget:
                    self._project.remove_widget(child_widget)

                self.object.remove_page(old_size - 1)
                old_size -= 1

prop_registry.override_property('GtkNotebook::n-pages', NPagesProp)

# Do not save property, since we need loader support since
# Gtk+ ignores the value of the property.
prop_registry.override_simple('GtkNotebook::page', persistent=False)




from django.utils.translation import ugettext_lazy as _

class CartAddProhibited(Exception):
    """Raised when a `signals.satchmo_cart_add_verify` listener vetoes adding an item to the cart.

    Params:
    - product: item which was being added
    - message: veto message
    """

    def __init__(self, product, message):
        self.product, self._message = product, message

    def _get_message(self):
        return self._message
    message = property(_get_message)

class OutOfStockError(CartAddProhibited):

    def __init__(self, product, have, need):
        if have == 0:
            msg = _("'%s' is out of stock.") % product.translated_name()
        else:
            msg = _("Only %(amount)i of '%(product)s' in stock.") % {
                'amount': have,
                'product': product.translated_name()
                }

        CartAddProhibited.__init__(self, product, msg)
        self.have = have
        self.need = need



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Self-contained demo script that bootstraps the helloworld
app and launch a WSGI server on localhost:8000. 

"""

import sys, os, logging
#Setup logging
log = logging.getLogger('notmm.controllers.wsgi')

from notmm.utils.django_settings import LazySettings
from notmm.utils.configparse import has_conf, loadconf
from notmm.http import get_bind_addr, HTTPServer

settings = LazySettings()

class ConfigurationError(Exception):
    pass

if has_conf(os.getcwd(), 'development.ini'):
    app_conf = loadconf('development.ini')
    # Get the host/port using get_bind_addr.
    bind_addr = get_bind_addr(app_conf, 'httpserver')
else:
    raise ConfigurationError('development.ini not found!')

def main(app_label):
    """Launch a simple WSGI app in stand-alone mode. 
    
    """
    # Setup basic site environment
    #configure_minimal_env()
    sys.path.extend(['lib', 'lib/site-packages'])

    if settings.DEBUG:
        log.debug("Found %d settings!"%settings.count())

    # setup default controller options
    from demo_app import WSGIApplication as WSGIApp
    wsgi_app = WSGIApp(settings=settings, app_conf=app_conf)
 
    # Install some handlers; We can skip this part if the
    # controller instance has them configured already.
    #wsgi_app.sethandle('handle404', app_label + '.handlers.handle404')
    #wsgi_app.sethandle('handle500', app_label + '.handlers.handle500')
    #import pdb; pdb.set_trace()
    WSGIServer = HTTPServer(wsgi_app, bind_addr)

    try:
        log.info("Starting HTTP server on %s:%d" % bind_addr)
        WSGIServer.serve()
    except (SystemExit, KeyboardInterrupt):
        log.info('Shutting down HTTP server...')
        sys.exit(2)

if __name__ == '__main__':
    main(app_label='helloworld')



import sys
#import mainapp
from notmm.controllers.wsgi import WSGIController
from notmm.controllers.auth import LoginController
from notmm.utils.wsgilib import HTTPRequest
from notmm.utils.configparse import loadconf

auth_conf = loadconf('development.ini', section='authkit')

from test_support import (
    WSGIControllerTestCase, 
    TestClient, 
    unittest,
    make_app,
    settings
    )

class AuthkitControllerTestCase(WSGIControllerTestCase):

    wsgi_app = WSGIController

    def setUp(self):
        self.test_init_middleware()
        self.client = TestClient(self.callback)

    def test_init_middleware(self):
        
        self.callback = LoginController(self.wsgi_app, auth_conf=auth_conf)
    

    def test_session_login(self):
        response = self.client.get('/session_login/') # 
        self.assertEqual(response[1], '200 OK')

    def test_session_login_POST(self):
        # XXX authenticate_user method test
        postdata = {
            'username': 'guest',
            'password': 'guest'}
        
        #client = TestClient(self.callback, method="POST")
        response = self.client.post('/session_login/', postdata)
        #req = HTTPRequest(environ=self.client.environ.copy())
        #self.assertEqual(req.method, 'POST')
        
        self.assertEqual(response[1], '200 OK')
        #self.assertEqual('REMOTE_USER' in response.environ, True)




from django.contrib import admin
from l10n.l10n_settings import get_l10n_setting
from product.modules.custom.models import CustomProduct, CustomTextField, CustomTextFieldTranslation

class CustomTextField_Inline(admin.TabularInline):
    model = CustomTextField
    extra = 3

class CustomTextFieldTranslation_Inline(admin.StackedInline):
    model = CustomTextFieldTranslation
    extra = 1

class CustomProductOptions(admin.ModelAdmin):
    inlines = [CustomTextField_Inline]

class CustomTextFieldOptions(admin.ModelAdmin):
    inlines = []
    if get_l10n_setting('show_admin_translations'):
        inlines.append(CustomTextFieldTranslation_Inline)

admin.site.register(CustomProduct, CustomProductOptions)
admin.site.register(CustomTextField, CustomTextFieldOptions)




#!/usr/bin/env python
from django.utils.translation.trans_real import DjangoTranslation
from django.middleware.locale            import LocaleMiddleware
from notmm.controllers.wsgi              import WSGIController
from notmm.utils.wsgilib                 import HTTPResponse
from test_support                        import *

class I18NController(WSGIController):

    i18n_middlewares = (LocaleMiddleware, )
    response_class = HTTPResponse

    #def get_locale_manager(self):
    #    return getattr(self, 'locale_manager')()
    
    def get_response(self, path_url):
        wsgi_response = super(self.__class__, self).get_response(path_url)
        for mw in self.i18n_middlewares:
            self.i18n_process_middleware(mw, 'process_request')
            self.i18n_process_middleware(mw, 'process_response', wsgi_response)
        return wsgi_response

    def i18n_process_middleware(self, mw, funcname, response=None):
            
        if self.request is not None:
            #print 'processing middleware: %s'%middleware
            if not response:
                getattr(mw(), funcname)(self.request)
            else:
                getattr(mw(), funcname)(self.request, response)
                # setup the WSGI response
                response.language_code = response.http_headers['Content-Language']
                # install the translations here

class I18NControllerTestCase(BaseControllerTestCase):
    wsgi_app = I18NController
    
    #def setUp(self):
    #    pass
    
    def test_get_current_locale(self):
        res = self.client.get('/')
        self.failUnlessEqual(res.status==200, True, res.status)
        lang_code = res.language_code
        self.assertEqual(lang_code=='en-us', True)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from notmm.utils.urlmap import RegexURLMap, url

handleLogin = 'blogengine2.contrib.session.views.login'
handleLogout = 'blogengine2.contrib.session.views.logout'
direct_to_template = 'notmm.utils.template.direct_to_template'

urlpatterns = RegexURLMap()
urlpatterns.add_routes('',
        # Frontdoor (a simple generic view)
        url(r'^$|index.html$', direct_to_template, {
            'template_name': 'sandbox/test.mako',
            #'charset': 'utf-8' # must specify mimetype or charset
        })
)

# Add the authentication views
urlpatterns.add_routes('',
        # Custom authentication backend wrapper (authkit based)
        # see nomainapp..controllers.auth.LoginController
        url(r'^session_login/$', handleLogin, \
            dict(template_name='auth/login.mako')),
        url(r'^session_logout/$', handleLogout, \
            dict(template_name='auth/logout.mako')))

urlpatterns.include('moviereviews.config.extras')



import os

MOUNTPOINTS = {
    'test' : dict(
        root=os.path.realpath(\
            os.path.join(\
                os.path.dirname(__file__), 'data')
                )
    ),
    'bob' : dict(
        root='/home/bob',
        username='bob',
        password='secret'
        )
}

# Use these settings, if config file does not define them (or is totally missing)
DEFAULT_CONFIG = {
    "mount_path": None,  # Application root, e.g. <mount_path>/<share_name>/<res_path>               
    "provider_mapping": {},
    "host": "localhost",
    "port": 8080, 
    "ext_servers": ["wsgiref", "cherrypy-bundled", "wsgidav"],
    "enable_loggers": [],
    "propsmanager": None,  # True: use property_manager.PropertyManager                  
    "locksmanager": True,  # True: use lock_manager.LockManager    
    
    # HTTP Authentication Options
    "user_mapping": {},       # dictionary of dictionaries 
    "domaincontroller": None, # None: domain_controller.WsgiDAVDomainController(user_mapping)
    "acceptbasic": True,      # Allow basic authentication, True or False
    "acceptdigest": True,     # Allow digest authentication, True or False
    "defaultdigest": True,    # True (default digest) or False (default basic)
    
    # Verbose Output
    "verbose": 3,       # 0 - no output (excepting application exceptions)         
                        # 1 - show single line request summaries (for HTTP logging)
                        # 2 - show additional events
                        # 3 - show full request/response header info (HTTP Logging)
                        #     request body and GET response bodies not shown
    
    "dir_browser": {
        "enable": True,          # Render HTML listing for GET requests on collections
        "response_trailer": "",  # Raw HTML code, appended as footer
        "davmount": False,       # Send <dm:mount> response if request URL contains '?davmount'
        "msmount": False,        # Add an 'open as webfolder' link (requires Windows)
    }
}




from decimal import Decimal
from livesettings import config_value

class Processor(object):
    
    method="percent"
    
    def __init__(self, order=None, user=None):
        """
        Any preprocessing steps should go here
        For instance, copying the shipping and billing areas
        """
        self.order = order
        self.user = user

    def by_orderitem(self, orderitem):
        if orderitem.product.taxable:
            price = orderitem.sub_total
            return self.by_price(orderitem.product.taxClass, price)
        else:
            return Decimal("0.00")
        
    def by_price(self, taxclass, price):
        percent = config_value('TAX','PERCENT')
        p = price * (percent/100)
        return p
        
    def by_product(self, product, quantity=Decimal('1')):
        price = product.get_qty_price(quantity)
        taxclass = product.taxClass
        return self.by_price(taxclass, price)
        
    def get_percent(self, *args, **kwargs):
        return Decimal(config_value('TAX','PERCENT'))
    
    def get_rate(self, *args, **kwargs):
        return self.get_rate_percent/100
        
    def shipping(self, subtotal=None):
        if subtotal is None and self.order:
            subtotal = self.order.shipping_sub_total

        if subtotal:
            subtotal = self.order.shipping_sub_total
            if config_value('TAX','TAX_SHIPPING'):
                percent = config_value('TAX','PERCENT')
                t = subtotal * (percent/100)
            else:
                t = Decimal("0.00")
        else:
            t = Decimal("0.00")
                
        return t
            
    def process(self, order=None):
        """
        Calculate the tax and return it
        """
        if order:
            self.order = order
        else:
            order = self.order

        percent = config_value('TAX','PERCENT')    

        sub_total = Decimal("0.00")
        for item in order.orderitem_set.filter(product__taxable=True):
            sub_total += item.sub_total
        
        itemtax = sub_total * (percent/100)
        taxrates = {'%i%%' % percent :  itemtax}
        
        if config_value('TAX','TAX_SHIPPING'):
            shipping = order.shipping_sub_total
            sub_total += shipping
            ship_tax = shipping * (percent/100)
            taxrates['Shipping'] = ship_tax
        
        tax = sub_total * (percent/100)
        return tax, taxrates



# Copyright (C) 2004,2005 by SICEm S.L., Imendio AB and Async Open Source
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import os
import sys
import warnings
import xml.dom.minidom
from xml.sax.saxutils import unescape

from kiwi.ui import dialogs
from kiwi.environ import environ

from gazpacho.config import get_app_data_dir
from gazpacho.library import load_library, LibraryLoadError
from gazpacho.loader.custom import valuefromstringsimpletypes, \
                                    get_pspec_from_name, str2bool
from gazpacho.properties import prop_registry
from gazpacho.widgetregistry import widget_registry

def xml_get_text_from_node(xmlnode):
    text = ''
    for node in xmlnode.childNodes:
        if node.nodeType == node.TEXT_NODE:
            text += node.data
    return unescape(text)

class BadCatalogSyntax(Exception):
    def __init__(self, file, msg):
        Exception.__init__(self, msg)
        self.file, self.msg = file, msg

    def __str__(self):
        return "Bad syntax in the catalog %s: %s" % (self.file, self.msg)

class CatalogError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)

class WidgetGroup(list):
    """List class used to store a widget group and its child nodes.

    The name property sets/gets the group internal name.
    The title property sets/gets the group title string.
    The class list contains the widget type strings for the group.
    """
    def __init__(self, name, title):
        list.__init__(self)
        self.name = name
        self.title = title

class Catalog:
    """Class to hold widget classes and groups.

    The title property gets/sets the catalog title string.
    The library property gets/sets the catalog library string.
    The widget_groups attribute holds a list of widget groups in the catalog.
    """
    def __init__(self, filename, resources_path):
        self.title = None
        self.library = None
        self.resource_path = resources_path
        self.widget_groups = []
        self.filename = filename
        self._parse_filename(filename)

    def _parse_filename(self, filename):
        fp = file(environ.find_resource('catalog', filename))
        document = xml.dom.minidom.parse(fp)
        root = document.documentElement
        self._parse_root(root)

    def _parse_root(self, node):
        # <!ELEMENT glade-catalog (glade-widget-classes?,
        #                          glade-widget-group*)>

        if node.nodeName == 'glade-catalog':
            self._parse_catalog(node)
        else:
            msg = "Root node is not glade catalog: %s" % node.nodeName
            raise BadCatalogSyntax(self.filename, msg)

        for child in node.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue

            if child.nodeName == 'glade-widget-classes':
                self._parse_widget_classes(child)
            elif child.nodeName == 'glade-widget-group':
                self._parse_widget_group(child)
            else:
                print('Unknown node: %s' % child.nodeName)

    def _parse_catalog(self, node):
        # <!ELEMENT glade-catalog (glade-widget-classes?,
        #                          glade-widget-group*)>
        # <!ATTLIST glade-catalog name            CDATA #REQUIRED
        #                         priority        CDATA #IMPLIED
        #                         library         CDATA #IMPLIED
        #                         library-type  (c|python) #IMPLIED
        #                         requires        CDATA #IMPLIED>

        # name
        self.title = str(node.getAttribute('name'))

        # priority
        priority = sys.maxsize
        if node.hasAttribute('priority'):
            priority = int(node.getAttribute('priority'))
        self.priority = priority

        # library
        library_name = str(node.getAttribute('library'))

        # library-type
        library_type = "python"
        if node.hasAttribute('library-type'):
            library_type = str(node.getAttribute('library-type'))

        name = os.path.basename(self.filename[:-4])
        try:
            self.library = load_library(library_type, name, library_name)
        except LibraryLoadError as e:
            raise CatalogError('Could not load library %s: %s' % (
                name, e))

    def _parse_widget_classes(self, node):
        "<!ELEMENT glade-widget-classes (glade-widget-class+)>"

        for class_node in node.childNodes:
            if class_node.nodeName != 'glade-widget-class':
                continue

            self._parse_widget_class(class_node)

    def _parse_widget_class(self, node):
        # <!ELEMENT glade-widget-class (tooltip?)>
        # <!ATTLIST glade-widget-class name          CDATA #REQUIRED
        #                              generic-name  CDATA #IMPLIED
	#	                       title         CDATA #IMPLIED
        #                              adaptor-class CDATA #IMPLIED>
        #                              register-type-function CDATA #IMPLIED>

        name = str(node.getAttribute('name'))
        if widget_registry.has_name(name):
            raise CatalogError("WidgetAdaptor %s is already registered" % name)

        generic_name = str(node.getAttribute('generic-name'))
        palette_name = str(node.getAttribute('title'))
        adaptor_class_name = str(node.getAttribute('adaptor-class'))

        tooltip = None
        for child in node.childNodes:
            if child.nodeName == 'tooltip':
                tooltip = xml_get_text_from_node(child)

        try:
            adaptor_class = self.library.get_widget_adaptor_class(
                name, adaptor_class_name)
        except LibraryLoadError as e:
            print('Could not load %s: %s' % (name, e))
            return

        adaptor = adaptor_class(name, generic_name, palette_name, self.library,
                                self.resource_path, tooltip)

        # store the Widget Adaptor on the cache
        widget_registry.add(adaptor)

    def _parse_widget_group(self, node):
        # <!ELEMENT glade-widget-group (glade-widget-class-ref+)>
        #  <!ATTLIST glade-widget-group name  CDATA #REQUIRED
        #                               title CDATA #REQUIRED>

        name = str(node.getAttribute('name'))
        title = str(node.getAttribute('title'))

        widget_group = WidgetGroup(name, title)

        for child in node.childNodes:
            if child.nodeName != 'glade-widget-class-ref':
                continue

            name = str(child.getAttribute('name'))
            widget_adaptor = widget_registry.get_by_name(name)
            if widget_adaptor is not None:
                widget_group.append(widget_adaptor)
            else:
                print('Could not load widget class %s' % name)

        self.widget_groups.append(widget_group)

    def __repr__(self):
        return '<Catalog %s>' % self.title

def _parse_gazpacho_path():
    """
    Parses the environment variable GAZPACHO_PATH and
    process all the directories in there
    """

    import gazpacho.widgets

    for path in os.environ.get('GAZPACHO_PATH', '').split(os.pathsep):
        if not path:
            continue

        environ.add_resource('pixmap', os.path.join(path, 'resources'))
        environ.add_resource('catalog', path)
        environ.add_resource('resource', os.path.join(path, 'resources'))
        environ.add_resource('glade', path)

        # library uses gazpacho.widgets as a base for plugins,
        # attach ourselves to it's __path__ so imports will work
        gazpacho.widgets.__path__.append(path)

def _get_catalog_dirs():
    catalog_error = ""

    _parse_gazpacho_path()
    catalogs = []
    for dirname in environ.get_resource_paths('catalog'):
        try:
            dir_list = os.listdir(dirname)
        except OSError as e:
            if e.errno == 2:
                catalog_error = ("The catalog directory %s does not exist" %
                                 dirname)
            elif e.errno == 13:
                catalog_error = ("You do not have read permissions in the "
                                 "catalog directory %s" % dirname)
            else:
                catalog_error = e.strerror

        if catalog_error:
            raise EnvironmentError(catalog_error)

        for filename in dir_list:
            if not filename.endswith('.xml'):
                continue
            name = filename[:-4]
            item = name, os.path.join(dirname, filename)
            # Make sure the base catalog is always loaded first, see #343283
            # We should introduce proper dependencies in the future.
            if name == 'base':
                catalogs.insert(0, item)
            else:
                catalogs.append(item)

    for name, filename in catalogs:
        if name == 'base':
            break
    else:
        raise SystemExit("Unable to find GTK+ catalog which is mandatory")

    return catalogs

_catalogs = []

def get_all_catalogs():
    """Return a list of existing Catalogs.

    The application widget registry is cleared and previously loaded classes
    and groups removed before rereading existing catalogs.
    The GTK+ catalog takes precedence and catalogs are ordered after
    their declared priority.
    """
    global _catalogs

    if not _catalogs:
        load_catalogs()

    assert _catalogs
    return _catalogs

def load_catalogs(external=True):
    """
    @param external: If true, allow loading of external (eg, not base) catalogs
    """
    try:
        catalogs = _get_catalog_dirs()
    except EnvironmentError as e:
        dialogs.error("<b>There were errors while loading Gazpacho</b>", str(e))
        raise SystemExit

    problems = []
    for name, filename in catalogs:
        # Skip external catalogs, for the testsuite.
        if not external and name != 'base':
            continue

        if filename.endswith('gtk+.xml'):
            warnings.warn("gtk+.xml found, it has been replaced by " +
                          "base.xml. Please remove it.", stacklevel=2)
            continue

        resources_path = environ.get_resource_paths('resource')
        try:
            catalog = Catalog(filename, resources_path)
            _catalogs.append(catalog)
        except IOError as e:
            problems.append(str(e))
        except BadCatalogSyntax as e:
            problems.append(str(e))
        except CatalogError as e:
            print('Error while loading catalog %s: %s' % (name, e))

    if problems:
        dialogs.error('\n'.join(problems))

    _catalogs.sort(lambda c1, c2: cmp(c1.priority,
                                      c2.priority))

# Default name of the file that contains custom UI definitions
CUSTOM_UI_FILENAME = 'custom.ui'

def get_custom_ui_filename():
    """Return the name of the file with custom UI definitions if it exists

    First it look in the user home directory (in the .gazpacho subdirectory),
    then it looks in every directory where there are catalogs.
    """
    # First we look in the user home directory
    app_dir = get_app_data_dir('gazpacho')
    custom_ui_filename = os.path.join(app_dir, CUSTOM_UI_FILENAME)
    if os.path.exists(custom_ui_filename):
        return custom_ui_filename

    # Then, look at standard catalog places
    for dirname in environ.get_resource_paths('catalog'):
        custom_ui_filename = os.path.join(dirname, CUSTOM_UI_FILENAME)
        if os.path.exists(custom_ui_filename):
            return custom_ui_filename

class CustomUI(object):
    """A Custom User Interface definition

    Using a XML file it is possible to redefine some parts of Gazpacho UI.
    Right now you can:

    - Change the palette grouping of widgets (you can mix widgets from
      different catalogs into the same palette groups)

    - Change the default value of a property of a widget

    - Hide a property making it non editable

    - Change the order of the properties in the editor by modifing its priority
    """

    def __init__(self, filename):
        """Constructor.

        It parses the file and changes the Gazpacho UI using its definitions
        """
        self.filename = filename
        self.widget_groups = []

        document = xml.dom.minidom.parse(file(filename))
        root = document.documentElement

        if root.nodeName != 'gazpacho-custom-ui':
            msg = "Root node is not gazpacho-custom-ui: %s" % root.nodeName
            raise BadCatalogSyntax(filename, msg)

        self._parse_root(root)

    def _parse_root(self, root):
        """Parser start point"""
        for child in root.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue

            if child.nodeName == 'palette-group':
                self._parse_custom_widget_group(child)
            else:
                print('Unknown node: %s' % child.nodeName)

    def _parse_custom_widget_group(self, node):
        """Parse a widget group and store it on a instance attribute"""
        name = str(node.getAttribute('name'))
        title = str(node.getAttribute('title'))

        widget_group = WidgetGroup(name, title)

        for child in node.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue

            if child.nodeName == 'widget':
                name = str(child.getAttribute('name'))
                widget_adaptor = widget_registry.get_by_name(name)
                if widget_adaptor is not None:
                    widget_group.append(widget_adaptor)
                    self._parse_custom_widget(name, child)
                else:
                    print('Could not load widget class %s' % name)

            else:
                print('Unknown node: %s' % child.nodeName)
        self.widget_groups.append(widget_group)

    def _parse_custom_widget(self, widget_name, node):
        """Parse a widget changing its property types"""
        for child in node.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue

            if child.nodeName == 'property':
                prop_name = str(child.getAttribute('name'))
                full_name = '%s::%s' % (widget_name, prop_name)
                attrs = {}
                for i in range(child.attributes.length):
                    attr = child.attributes.item(i)
                    name = str(attr.name)
                    value = attr.value
                    if name == 'name':
                        continue

                    if name == 'default':
                        value = _valuefromstring(widget_name, prop_name, value)

                    elif name in ('readable', 'writable', 'enabled',
                                  'translatable', 'editable'):
                        value = str2bool(value)

                    elif name == 'priority':
                        value = int(value)

                    else:
                        print('Customization of attribute %s is not supported' % name)
                        continue

                    attrs[str(attr.name)] = value

                if 'default' in list(attrs.keys()):
                    attrs['has_custom_default'] = True

                prop_registry.override_simple(full_name, **attrs)

            else:
                print('Unknown node: %s' % child.nodeName)

def _valuefromstring(widget_name, prop_name, string_value):
    """Helper function to convert a string to a valid value

    This is used when parsing default values for custom ui properties.
    """
    pspec = get_pspec_from_name(widget_name, prop_name)
    prop_type = pspec.value_type
    value = valuefromstringsimpletypes(prop_type, string_value)
    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Database opener middleware for WSGI apps.

 Copyright (C) 2001-2007 Orbtech, L.L.C.
 Copyright (C) 2009-2019 Etienne Robillard <tkadm30@yandex.com>

 libschevo - Another reimplementation of the Schevo DBMS.
 https://isotopesoftware.ca/software/libschevo/

 This toolkit is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This toolkit is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 For copyright, license, and warranty, see bottom of file.

"""

from notmm.controllers.wsgi import WSGIController
from notmm.dbapi.orm.schevo_compat import XdserverProxy

__all__ = ['SchevoController']

class SchevoController(WSGIController):
    """Schevo Database connection middleware.

    A SchevoController instance will look for configuration items whose
    keys begin with ``schevo.db.``, and open the corresponding
    database using the location given by that configuration item.

    During WSGI application invocation, a SchevoController attaches
    those same keys in the WSGI environment, with the values being the
    open Schevo databases.  It also attaches a key ``schevo.db``,
    which is a dictionary whose keys are the names of each database and
    whose values are the open Schevo databases.

    The SchevoController instance itself is injected into the WSGI
    environment with the key ``schevo.dbopener``.  This allows other
    components in the WSGI stack to call the `open` and `close`
    methods to open and close more databases, respectively.
    """

    key_prefix = 'schevo.db.'
    backendClass = XdserverProxy

    def __init__(self, request, db_name, **kwargs):
        """Create a new ``SchevoController`` instance.

        request: The request to filter.
        """
        super(SchevoController, self).__init__(**kwargs)

        self.db_name = db_name
        self.environ_key = self.key_prefix + db_name
        self.setup_database(db_name)

        # Enable logging if debug is set
        #if (self.logger and self.debug):
        #    #self.logger.debug(repr(self))
        #    self.logger.debug("Configured schevo database: %s" % self.environ_key)

    def setup_database(self, db_name):
        # Adds a db instance to the current request object (self)
        self.db = self.backendClass(db_name)

    def __repr__(self):
        return "<SchevoController: %s>" % self.db_name

    def init_request(self, environ):
        super(SchevoController, self).init_request(environ)
        self.request.environ[self.environ_key] = self.db



from shipping.modules.tieredweight.models import Carrier, Shipper
from satchmo_utils import load_once

load_once('tiered', 'shipping.modules.tieredweight')

def get_methods():
    return [Shipper(carrier) for carrier in Carrier.objects.filter(active=True)]



from django import template
from django.conf import settings
from django.core import urlresolvers
from django.template import Context, Template
from django.utils.translation import get_language, ugettext_lazy as _
from livesettings import config_value
from product.models import Category, Product
from product.queries import bestsellers
from satchmo_utils.templatetags import get_filter_args
import keyedcache

register = template.Library()

@register.filter
def best_selling_products_list(count):
    """Get a list of best selling products"""
    try:
        ct = int(count)
    except ValueError:
        ct = config_value('PRODUCT','NUM_PAGINATED')
    
    return bestsellers(ct)

@register.filter
def recent_products_list(count):
    """Get a list of recent products"""
    try:
        ct = int(count)
    except ValueError:
        ct = config_value('PRODUCT','NUM_PAGINATED')
     
    query = Product.objects.recent_by_site()
    return query[:ct]

def is_producttype(product, ptype):
    """Returns True if product is ptype"""
    if ptype in product.get_subtypes():
        return True
    else:
        return False

register.filter('is_producttype', is_producttype)

def product_count(category, args=''):
    """Get a count of products for the base object.
    
    If `category` is None, then count everything.
    If it is a `Category` object then count everything in the category and subcategories.
    """
    args, kwargs = get_filter_args(args, boolargs=('variations'))
    variations = kwargs.get('variations', False)
    try:
        ct = keyedcache.cache_get('product_count', category, variations)
    except keyedcache.NotCachedError:
        if not category:
            ct = Product.objects.active_by_site(variations=variations).count()
        else:
            ct = category.active_products(include_children=True, variations=variations).count()
            
        keyedcache.cache_set('product_count', category, args, value=ct)
    return ct
    
register.filter('product_count', product_count)

def product_images(product, args=""):
    args, kwargs = get_filter_args(args,
        keywords=('include_main', 'maximum'),
        boolargs=('include_main'),
        intargs=('maximum'),
        stripquotes=True)

    q = product.productimage_set
    if kwargs.get('include_main', True):
        q = q.all()
    else:
        main = product.main_image
        q = q.exclude(id = main.id)

    maximum = kwargs.get('maximum', -1)
    if maximum>-1:
        q = list(q)[:maximum]

    return q

register.filter('product_images', product_images)

def smart_attr(product, key):
    """
    Run the smart_attr function on the spec'd product
    """
    return product.smart_attr(key)

register.filter('smart_attr', smart_attr)


def product_sort_by_price(products):
    """
    Sort a product list by unit price
    
    Example::
        
        {% for product in products|product_sort_by_price %} 
    """
    
    if products:
        fast = [(product.unit_price, product) for product in products]
        fast.sort()
        return zip(*fast)[1]
    
register.filter('product_sort_by_price', product_sort_by_price)



from django.template import Context, loader
from payment.modules.base import BasePaymentProcessor, ProcessorResult, NOTSET
from satchmo_utils.numbers import trunc_decimal
from django.utils.translation import ugettext_lazy as _

import urllib2
try:
    from xml.etree.ElementTree import fromstring
except ImportError:
    from elementtree.ElementTree import fromstring

# Response codes available at:
# http://apps.cybersource.com/library/documentation/sbc/api_guide/SB_API.pdf
CYBERSOURCE_RESPONSES = {
    '100' : 'Successful transaction.',
    '101' : 'The request is missing one or more required fields.',
    '102' : 'One or more fields in the request contains invalid data.',
    '104' : 'The merchantReferenceCode sent with this authorization request matches the merchantReferenceCode of another authorization request that you sent in the last 15 minutes.',
    '150' : 'Error: General system failure. ',
    '151' : 'Error: The request was received but there was a server timeout. This error does not include timeouts between the client and the server.',
    '152' : 'Error: The request was received, but a service did not finish running in time.',
    '201' : 'The issuing bank has questions about the request. You do not receive an authorization code in the reply message, but you might receive one verbally by calling the processor.',
    '202' : 'Expired card. You might also receive this if the expiration date you provided does not match the date the issuing bank has on file.',
    '203' : 'General decline of the card. No other information provided by the issuing bank.',
    '204' : 'Insufficient funds in the account.',
    '205' : 'Stolen or lost card.',
    '207' : 'Issuing bank unavailable.',
    '208' : 'Inactive card or card not authorized for card-not-present transactions.',
    '210' : 'The card has reached the credit limit. ',
    '211' : 'Invalid card verification number.',
    '221' : 'The customer matched an entry on the processor\'s negative file.',
    '231' : 'Invalid account number.',
    '232' : 'The card type is not accepted by the payment processor.',
    '233' : 'General decline by the processor.',
    '234' : 'There is a problem with your CyberSource merchant configuration.',
    '235' : 'The requested amount exceeds the originally authorized amount. Occurs, for example, if you try to capture an amount larger than the original authorization amount. This reason code only applies if you are processing a capture through the API.',
    '236' : 'Processor Failure',
    '238' : 'The authorization has already been captured. This reason code only applies if you are processing a capture through the API.',
    '239' : 'The requested transaction amount must match the previous transaction amount. This reason code only applies if you are processing a capture or credit through the API.',
    '240' : 'The card type sent is invalid or does not correlate with the credit card number.',
    '241' : 'The request ID is invalid. This reason code only applies when you are processing a capture or credit through the API.',
    '242' : 'You requested a capture through the API, but there is no corresponding, unused authorization record. Occurs if there was not a previously successful authorization request or if the previously successful authorization has already been used by another capture request. This reason code only applies when you are processing a capture through the API.',
    '243' : 'The transaction has already been settled or reversed.',
    '246' : 'The capture or credit is not voidable because the capture or credit information has already been submitted to your processor. Or, you requested a void for a type of transaction that cannot be voided. This reason code applies only if you are processing a void through the API.',
    '247' : 'You requested a credit for a capture that was previously voided. This reason code applies only if you are processing a void through the API.',
    '250' : 'Error: The request was received, but there was a timeout at the payment processor.',
    '520' : 'The authorization request was approved by the issuing bank but declined by CyberSource based on your Smart Authorization settings.',
}


class PaymentProcessor(BasePaymentProcessor):
    """
    Cybersource payment processing module
    You must have an account with Cybersource in order to use this module
    
    """
    def __init__(self, settings):
        super(PaymentProcessor, self).__init__('cybersource', settings)
        self.contents = ''
        if settings.LIVE.value:
            self.testflag = 'FALSE'
            self.connection = settings.CONNECTION.value
        else:
            self.testflag = 'TRUE'
            self.connection = settings.CONNECTION_TEST.value
            
        self.configuration = {
            'merchantID' : settings.MERCHANT_ID.value,
            'password' : settings.TRANKEY.value,
            }

    def prepare_content(self, order, amount):
        self.bill_to = {
            'firstName' : order.contact.first_name,
            'lastName' : order.contact.last_name,
            'street1': order.full_bill_street,
            'city': order.bill_city,
            'state' : order.bill_state,
            'postalCode' : order.bill_postal_code,
            'country': order.bill_country,
            'email' : order.contact.email,
            'phoneNumber' : order.contact.primary_phone,
            # Can add additional info here if you want to but it's not required
            }
        exp = order.credit_card.expirationDate.split('/')
        self.card = {
            'accountNumber' : order.credit_card.decryptedCC,
            'expirationMonth' : exp[0],
            'expirationYear' : exp[1],
            'cvNumber' : order.credit_card.ccv
            }
        currency = self.settings.CURRENCY_CODE.value
        currency = currency.replace("_", "")
        self.purchase_totals = {
            'currency' : currency,
            'grandTotalAmount' : trunc_decimal(amount, 2),
        }

    def capture_payment(self, testing=False, order=None, amount=NOTSET):
        """
        Creates and sends XML representation of transaction to Cybersource
        """
        if not order:
            order = self.order
            
        if order.paid_in_full:
            self.log_extra('%s is paid in full, no capture attempted.', order)
            self.record_payment()
            return ProcessorResult(self.key, True, _("No charge needed, paid in full."))

        self.log_extra('Capturing payment for %s', order)

        if amount==NOTSET:
            amount = order.balance

        self.prepare_content(order, amount)
        
        invoice = "%s" % order.id
        failct = order.paymentfailures.count()
        if failct > 0:
            invoice = "%s_%i" % (invoice, failct)
        
        # XML format is very simple, using ElementTree for generation would be overkill
        t = loader.get_template('shop/checkout/cybersource/request.xml')
        c = Context({
            'config' : self.configuration,
            'merchantReferenceCode' : invoice,
            'billTo' : self.bill_to,
            'purchaseTotals' : self.purchase_totals,
            'card' : self.card,
        })
        request = t.render(c)
        conn = urllib2.Request(url=self.connection, data=request)
        try:
            f = urllib2.urlopen(conn)
        except urllib2.HTTPError, e:
            # we probably didn't authenticate properly
            # make sure the 'v' in your account number is lowercase
            return ProcessorResult(self.key, False, 'Problem parsing results')

        f = urllib2.urlopen(conn)
        all_results = f.read()
        tree = fromstring(all_results)
        parsed_results = tree.getiterator('{urn:schemas-cybersource-com:transaction-data-1.26}reasonCode')
        try:
            reason_code = parsed_results[0].text
        except KeyError:
            return ProcessorResult(self.key, False, 'Problem parsing results')

        response_text = CYBERSOURCE_RESPONSES.get(reason_code, 'Unknown Failure')

        if reason_code == '100':
            payment = self.record_payment(order=order, amount=amount, 
                transaction_id="", reason_code=reason_code)
            return ProcessorResult(self.key, True, response_text, payment=payment)
        else:
            payment = self.record_failure(order=order, amount=amount, 
                transaction_id="", reason_code=reason_code, 
                details=response_text)
            
            return ProcessorResult(self.key, False, response_text)

if __name__ == "__main__":
    """
    For testing purposes only.
    Allows this module to be run as a script to test the connection
    
    """

    import os
    from livesettings import config_get_group

    # Set up some dummy classes to mimic classes being passed through Satchmo
    class testContact(object):
        pass
    class testCC(object):
        pass
    class testOrder(object):
        def __init__(self):
            self.contact = testContact()
            self.credit_card = testCC()
        def order_success(self):
            pass

    if not os.environ.has_key("DJANGO_SETTINGS_MODULE"):
        os.environ["DJANGO_SETTINGS_MODULE"]="satchmo_store.settings"

    settings_module = os.environ['DJANGO_SETTINGS_MODULE']
    settingsl = settings_module.split('.')
    settings = __import__(settings_module, {}, {}, settingsl[-1])

    sampleOrder = testOrder()
    sampleOrder.id = '1234'
    sampleOrder.contact.first_name = 'Chris'
    sampleOrder.contact.last_name = 'Smith'
    sampleOrder.contact.primary_phone = '801-555-9242'
    sampleOrder.contact.email = 'null@cybersource.com'
    sampleOrder.full_bill_street = '123 Main Street'
    sampleOrder.bill_postal_code = '12345'
    sampleOrder.bill_state = 'TN'
    sampleOrder.bill_city = 'Some City'
    sampleOrder.bill_country = 'US'
    sampleOrder.total = "27.00"
    sampleOrder.balance = "27.00"
    sampleOrder.credit_card.decryptedCC = '6011000000000012'
    sampleOrder.credit_card.expirationDate = "10/09"
    sampleOrder.credit_card.ccv = "144"

    cybersource_settings = config_get_group('PAYMENT_CYBERSOURCE')
    if cybersource_settings.LIVE.value:
        print "Warning.  You are submitting a live order.  CYBERSOURCE system is set LIVE."
        
    processor = PaymentProcessor(cybersource_settings)
    processor.prepare_data(sampleOrder)
    results = processor.process(testing=True)
    print results





# -*- coding: utf-8 -*-
"""__init__ module for the threadlocals package

:Authors:
    - Herbert Poul http://sct.sphene.net
    - Bruce Kroeze http://gosatchmo.com
"""
"""
New BSD License
===============
Copyright (c) 2008, Bruce Kroeze http://solidsitesolutions.com

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of SolidSiteSolutions LLC, Zefamily LLC nor the names of its 
      contributors may be used to endorse or promote products derived from this 
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
__docformat__="restructuredtext"

import logging

log = logging.getLogger('ThreadLocalMiddleware')

try:
    from threading import local
except ImportError:
    log.warn('getting threadlocal support from django')
    from django.utils._threading_local import local
    
from django.contrib.sites.models import Site

_threadlocals = local()

def get_current_host(request = None):
    if not request:
        request = get_current_request()
    if request:
        try:
            host = request.META["HTTP_HOST"]
            return host
        except KeyError:
            pass
    #log.warn('No request found - returning default site')
    site = Site.objects.get_current()
    return site.domain

def get_current_request():
    return get_thread_variable('request', None)

def get_current_session():
    req = get_current_request()
    if req == None: return None
    return req.session
    
def get_current_user():
    user = get_thread_variable('user', None)
    if user != None: return user
    req = get_current_request()
    if req == None: return None
    return req.user

def set_current_user(user):
    set_thread_variable('user', user)

def set_thread_variable(key, var):
    setattr(_threadlocals, key, var)
    
def get_thread_variable(key, default=None):
    return getattr(_threadlocals, key, default)


'''
Associable Elixir Statement Generator

==========
Associable
==========

About Polymorphic Associations
------------------------------

A frequent pattern in database schemas is the has_and_belongs_to_many, or a
many-to-many table. Quite often multiple tables will refer to a single one
creating quite a few many-to-many intermediate tables.

Polymorphic associations lower the amount of many-to-many tables by setting up
a table that allows relations to any other table in the database, and relates
it to the associable table. In some implementations, this layout does not
enforce referential integrity with database foreign key constraints, this
implementation uses an additional many-to-many table with foreign key
constraints to avoid this problem.

.. note:
    SQLite does not support foreign key constraints, so referential integrity
    can only be enforced using database backends with such support.

Elixir Statement Generator for Polymorphic Associations
-------------------------------------------------------

The ``associable`` function generates the intermediary tables for an Elixir
entity that should be associable with other Elixir entities and returns an
Elixir Statement for use with them. This automates the process of creating the
polymorphic association tables and ensuring their referential integrity.

Matching select_XXX and select_by_XXX are also added to the associated entity
which allow queries to be run for the associated objects.

Example usage:

.. sourcecode:: python

    class Tag(Entity):
        name = Field(Unicode)

    acts_as_taggable = associable(Tag)

    class Entry(Entity):
        title = Field(Unicode)
        acts_as_taggable('tags')

    class Article(Entity):
        title = Field(Unicode)
        acts_as_taggable('tags')

Or if one of the entities being associated should only have a single member of
the associated table:

.. sourcecode:: python

    class Address(Entity):
        street = Field(String(130))
        city = Field(String(100))

    is_addressable = associable(Address, 'addresses')

    class Person(Entity):
        name = Field(Unicode)
        orders = OneToMany('Order')
        is_addressable()

    class Order(Entity):
        order_num = Field(primary_key=True)
        item_count = Field(Integer)
        person = ManyToOne('Person')
        is_addressable('address', uselist=False)

    home = Address(street='123 Elm St.', city='Spooksville')
    user = Person(name='Jane Doe')
    user.addresses.append(home)

    neworder = Order(item_count=4)
    neworder.address = home
    user.orders.append(neworder)

    # Queries using the added helpers
    Person.select_by_addresses(city='Cupertino')
    Person.select_addresses(and_(Address.c.street=='132 Elm St',
                                 Address.c.city=='Smallville'))

Statement Options
-----------------

The generated Elixir Statement has several options available:

+---------------+-------------------------------------------------------------+
| Option Name   | Description                                                 |
+===============+=============================================================+
| ``name``      | Specify a custom name for the Entity attribute. This is     |
|               | used to declare the attribute used to access the associated |
|               | table values. Otherwise, the name will use the plural_name  |
|               | provided to the associable call.                            |
+---------------+-------------------------------------------------------------+
| ``uselist``   | Whether or not the associated table should be represented   |
|               | as a list, or a single property. It should be set to False  |
|               | when the entity should only have a single associated        |
|               | entity. Defaults to True.                                   |
+---------------+-------------------------------------------------------------+
| ``lazy``      | Determines eager loading of the associated entity objects.  |
|               | Defaults to False, to indicate that they should not be      |
|               | lazily loaded.                                              |
+---------------+-------------------------------------------------------------+
'''
from elixir.statements import Statement
import sqlalchemy as sa

__doc_all__ = ['associable']


def associable(assoc_entity, plural_name=None, lazy=True):
    '''
    Generate an associable Elixir Statement
    '''
    interface_name = assoc_entity._descriptor.tablename
    able_name = interface_name + 'able'

    if plural_name:
        attr_name = "%s_rel" % plural_name
    else:
        plural_name = interface_name
        attr_name = "%s_rel" % interface_name

    class GenericAssoc(object):

        def __init__(self, tablename):
            self.type = tablename

    #TODO: inherit from entity builder
    class Associable(object):
        """An associable Elixir Statement object"""

        def __init__(self, entity, name=None, uselist=True, lazy=True):
            self.entity = entity
            self.lazy = lazy
            self.uselist = uselist

            if name is None:
                self.name = plural_name
            else:
                self.name = name

        def after_table(self):
            col = sa.Column('%s_assoc_id' % interface_name, sa.Integer,
                            sa.ForeignKey('%s.id' % able_name))
            self.entity._descriptor.add_column(col)

            if not hasattr(assoc_entity, '_assoc_table'):
                metadata = assoc_entity._descriptor.metadata
                association_table = sa.Table("%s" % able_name, metadata,
                    sa.Column('id', sa.Integer, primary_key=True),
                    sa.Column('type', sa.String(40), nullable=False),
                )
                tablename =  "%s_to_%s" % (able_name, interface_name)
                association_to_table = sa.Table(tablename, metadata,
                    sa.Column('assoc_id', sa.Integer,
                              sa.ForeignKey(association_table.c.id,
                                            ondelete="CASCADE"),
                              primary_key=True),
                    #FIXME: this assumes a single id col
                    sa.Column('%s_id' % interface_name, sa.Integer,
                              sa.ForeignKey(assoc_entity.table.c.id,
                                            ondelete="RESTRICT"),
                              primary_key=True),
                )

                assoc_entity._assoc_table = association_table
                assoc_entity._assoc_to_table = association_to_table

        def after_mapper(self):
            if not hasattr(assoc_entity, '_assoc_mapper'):
                assoc_entity._assoc_mapper = sa.orm.mapper(
                    GenericAssoc, assoc_entity._assoc_table, properties={
                        'targets': sa.orm.relation(
                                       assoc_entity,
                                       secondary=assoc_entity._assoc_to_table,
                                       lazy=lazy, backref='associations',
                                       order_by=assoc_entity.mapper.order_by)
                })

            entity = self.entity
            entity.mapper.add_property(
                attr_name,
                sa.orm.relation(GenericAssoc, lazy=self.lazy,
                                backref='_backref_%s' % entity.table.name)
            )

            if self.uselist:
                def get(self):
                    if getattr(self, attr_name) is None:
                        setattr(self, attr_name,
                                GenericAssoc(entity.table.name))
                    return getattr(self, attr_name).targets
                setattr(entity, self.name, property(get))
            else:
                # scalar based property decorator
                def get(self):
                    attr = getattr(self, attr_name)
                    if attr is not None:
                        return attr.targets[0]
                    else:
                        return None
                def set(self, value):
                    if getattr(self, attr_name) is None:
                        setattr(self, attr_name,
                                GenericAssoc(entity.table.name))
                    getattr(self, attr_name).targets = [value]
                setattr(entity, self.name, property(get, set))

            # self.name is both set via mapper synonym and the python
            # property, but that's how synonym properties work.
            # adding synonym property after "real" property otherwise it
            # breaks when using SQLAlchemy > 0.4.1
            entity.mapper.add_property(self.name, sa.orm.synonym(attr_name))

            # add helper methods
            def select_by(cls, **kwargs):
                return cls.query.join([attr_name, 'targets']) \
                                .filter_by(**kwargs).all()
            setattr(entity, 'select_by_%s' % self.name, classmethod(select_by))

            def select(cls, *args, **kwargs):
                return cls.query.join([attr_name, 'targets']) \
                                .filter(*args, **kwargs).all()
            setattr(entity, 'select_%s' % self.name, classmethod(select))

    return Statement(Associable)



# Copyright (C) 2005 by Async Open Source
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gtk

from kiwi.environ import environ

# XXX: Move to a file
MAIN_UI_STRING = """
<ui>
  <menubar name="MainMenu">
    <menu action="FileMenu">
      <menuitem action="New"/>
      <menuitem action="Open"/>
      <separator name="FM1"/>
      <menuitem action="Save"/>
      <menuitem action="SaveAs"/>
      <separator name="FM2"/>
      <placeholder name="RecentProjects"/>
      <separator name="FM3"/>
      <menuitem action="Close"/>
      <menuitem action="Quit"/>
    </menu>
    <menu action="EditMenu">
      <menuitem action="Undo"/>
      <menuitem action="Redo"/>
      <separator name="EM1"/>
      <menuitem action="Cut"/>
      <menuitem action="Copy"/>
      <menuitem action="Paste"/>
      <menuitem action="Delete"/>
      <separator name="EM2"/>
      <menuitem action="ShowStructure"/>
      <menuitem action="ShowWorkspace"/>
      <separator name="EM3"/>
      <menuitem action="Preferences"/>
    </menu>
    <menu action="ObjectMenu">
    </menu>
    <menu action="ProjectMenu">
      <placeholder name="OpenProjects"/>
      <separator name="project-separtor"/>
      <menuitem action="ProjectProperties"/>
    </menu>
    <menu action="DebugMenu">
      <menuitem action="ShowCommandStack"/>
      <menuitem action="ShowClipboard"/>
      <separator/>
      <menuitem action="Preview"/>
      <menuitem action="DumpData"/>
    </menu>
    <menu action="HelpMenu">
      <menuitem action="About"/>
    </menu>
  </menubar>
  <toolbar name="MainToolbar">
    <toolitem action="Open"/>
    <toolitem action="Save"/>
    <separator name="MT1"/>
    <toolitem action="Undo"/>
    <toolitem action="Redo"/>
    <separator name="MT2"/>
    <toolitem action="Cut"/>
    <toolitem action="Copy"/>
    <toolitem action="Paste"/>
    <separator name="MT3"/>
  </toolbar>
</ui>
"""

class BarManager(object):
    """
    Can I have a beer with that please?
    """

    def __init__(self):
        # Mapping of action group name -> ActionGroup
        self._action_groups = {}
        # Mapping of action name -> Action
        self._action_cache = {}

        self._ui_manager = gtk.UIManager()

        # Create action groups
        for group_name in ['Normal', 'ContextActions',
                           'OpenProjects',
                           'RecentProjects']:
            self._create_action_group(group_name)
        self._create_action_group('AlwaysDisabled', False)

    def build_interfaces(self):
        self._ui_manager.add_ui_from_string(MAIN_UI_STRING)

        # Make sure that the project and object menu isn't removed if empty
        self.set_action_prop('ProjectMenu', hide_if_empty=False)
        self.set_action_prop('ObjectMenu', hide_if_empty=False)

        toolbar = self.get_toolbar()
        toolbar.set_style(gtk.TOOLBAR_ICONS)

        # Disable workspace for older versions of PyGTK.
        if gtk.pygtk_version < (2, 8):
            self.set_action_prop('ShowWorkspace', sensitive=False)

    def _create_action_group(self, name, sensitive=True):
        action_group = gtk.ActionGroup(name)
        action_group.set_sensitive(sensitive)
        self.add_action_group(action_group)
        return action_group

    def add_action_group(self, action_group):
        self._action_groups[action_group.get_name()] = action_group
        self._ui_manager.insert_action_group(action_group, 0)
        for action in action_group.list_actions():
            self._action_cache[action.get_name()] = action

    def remove_action_group(self, name):
        action_group = self._action_groups[name]
        del self._action_groups[name]
        self._ui_manager.remove_action_group(action_group)
        for action in action_group.list_actions():
            del self._action_cache[action.get_name()]

    def get_group(self, group_name):
        """
        @param group_name: name of an action group
        """
        if not group_name in self._action_groups:
            raise ValueError("There is no action group called %s" % group_name)
        return self._action_groups[group_name]

    def has_group(self, group_name):
        """
        @param group_name: name of an action group
        """
        return group_name in self._action_groups

    def add_action(self, group_name, action):
        """
        @param group_name: name of an action group
        @param action: a gtk.Action
        """
        if not isinstance(action, gtk.Action):
            raise TypeError("action must be a gtk.Action")

        action_group = self.get_group(group_name)

        action_name = action.get_name()
        if action_name in self._action_cache:
            action_group.remove_action(action)

        action_group.add_action(action)
        self._action_cache[action_name] = action

    def add_actions(self, group_name, *actions):
        """
        @param group_name: name of an action group
        @param actions: sequence of action tuples
        """
        action_group = self.get_group(group_name)
        action_group.add_actions(actions)
        for action in actions:
            name = action[0]
            action = action_group.get_action(name)
            self._action_cache[name] = action

    def add_toggle_actions(self, group_name, *actions):
        """
        @param group_name: name of an action group
        @param actions: sequence of toggle action tuples
        """
        action_group = self.get_group(group_name)
        action_group.add_toggle_actions(actions)
        for action in actions:
            name = action[0]
            action = action_group.get_action(name)
            self._action_cache[name] = action

    def set_action_prop(self, action_name, **properties):
        """
        Gets an action and sets properties of it.
        Note that action needs to be added through add_action,
        add_actions or add_toggle_actions in this class.
        @param action_name: name of an action
        @param properties: properties to set
        """
        action = self._get_action(action_name)
        for key, value in list(properties.items()):
            action.set_property(key, value)

    def set_action_props(self, action_names, **properties):
        """
        An extended version of set_action_props, which takes a sequence
        of action_names instead of just a name of one action
        @param action_names: a sequence of names of actions
        @param properties: properties to set
        """
        if not isinstance(action_names, (list, tuple)):
            raise TypeError("action_names must be a list or tuple")

        for action_name in action_names:
            self.set_action_prop(action_name, **properties)

    def activate_action(self, name):
        """
        Activate the specified action.

        @param name: the action name
        @type name: str
        """
        action = self._get_action(name)
        action.activate()

    def _get_action(self, action_name):
        """
        Get the first occurence of an action with the given name.

        @param action_name: the name of the action
        @type action_name: str

        @return: the action
        @rtype: gtk.Action
        """
        if not action_name in self._action_cache:
            raise ValueError("No action called: %s" % action_name)

        return self._action_cache[action_name]


    # GtkUIManager delegators

    def get_menubar(self):
        return self._ui_manager.get_widget('/MainMenu')

    def get_toolbar(self):
        return self._ui_manager.get_widget('/MainToolbar')

    def get_accel_group(self):
        return self._ui_manager.get_accel_group()

    def get_action(self, action_name):
        return self._ui_manager.get_action(action_name)

    def get_widget(self, widget_name):
        return self._ui_manager.get_widget(widget_name)

    def remove_ui(self, uid):
        self._ui_manager.remove_ui(uid)

    def add_ui_from_string(self, ui_string):
        return self._ui_manager.add_ui_from_string(ui_string)

    def ensure_update(self):
        self._ui_manager.ensure_update()

if environ.epydoc:
    bar_manager = object()
else:
    bar_manager = BarManager()
